import unittest
from collections import OrderedDict
import shutil
from kojen.preservative import *

class TestPreservative(unittest.TestCase):
    """ Unit test. Not sure how to programmatically test this.
        Its better done by inspection?
    """
    def setUp(self):
        self.TempFolder = "testpreservative"
        self.HeaderFileName = os.path.join(self.TempFolder, "TestPreservative.h")
        self.SourceFileName = os.path.join(self.TempFolder, "TestPreservative.cpp")
        self.Preserve1 = "aabbccddeeffgghh"
        self.Preserve2 = "1122334455667788"
        self.Preserve3 = "howdydooody"
        self.removeDir()
        self.makeDir()
        self.makeFile(self.HeaderFileName)
        self.makeFile(self.SourceFileName)

    def tearDown(self):
        self.removeDir()

    def removeDir(self):
        if os.path.exists(self.TempFolder):
            try:
                shutil.rmtree(self.TempFolder)
            except OSError:
                print("Deletion of the directory %s failed" % self.TempFolder)

    def makeDir(self):
        try:
            os.mkdir(self.TempFolder)
        except OSError:
            print("Creation of the directory %s failed" % self.TempFolder)

    def makeFile(self, filename):
        with open(filename, 'w+') as temp_file:
            for i in range(10):
                temp_file.write('1' + i*"1" + '\n')
            temp_file.write('{{{USER_STUFF_1\n')
            temp_file.write(self.Preserve1 + '\n')
            temp_file.write(self.Preserve1 + '\n')
            temp_file.write(self.Preserve1 + '\n')
            temp_file.write('{{{USER_STUFF_1\n')
            for i in range(20):
                temp_file.write(10*'2' + i*"2" + '\n')
            temp_file.write('{{{USER_STUFF_2\n')
            temp_file.write(self.Preserve2 + '\n')
            temp_file.write(self.Preserve2 + '\n')
            temp_file.write(self.Preserve2 + '\n')
            temp_file.write('{{{USER_STUFF_2\n')
            for i in range(30):
                temp_file.write(20*'3' + i*"3" + '\n')
            temp_file.write('{{{USER_STUFF_3\n')
            temp_file.write(self.Preserve3 + '\n')
            temp_file.write(self.Preserve3 + '\n')
            temp_file.write(self.Preserve3 + '\n')
            temp_file.write('{{{USER_STUFF_3\n')

    def checkOutput(self, ordered_dictionary_of_outputs):
        for name, f in ordered_dictionary_of_outputs.items():
            user1 = False
            user2 = False
            user3 = False
            self.assertIn("{{{USER_STUFF_1", f, "Missing tag (1) in output")
            self.assertIn("{{{USER_STUFF_2", f, "Missing tag (2) in output")
            self.assertIn("{{{USER_STUFF_3", f, "Missing tag (3) in output")
            for i in range(len(f)):
                line = f[i]
                if not user1 and "{{{USER_STUFF_1" in line:
                    user1 = True
                    self.assertTrue(f[i + 1].find(self.Preserve1) != -1)
                    self.assertTrue(f[i + 2].find(self.Preserve1) != -1)
                    self.assertTrue(f[i + 3].find(self.Preserve1) != -1)
                    self.assertTrue(f[i + 4].find("{{{USER_STUFF_1") != -1)
                if not user2 and "{{{USER_STUFF_2" in line:
                    user2 = True
                    self.assertTrue(f[i + 1].find(self.Preserve2) != -1)
                    self.assertTrue(f[i + 2].find(self.Preserve2) != -1)
                    self.assertTrue(f[i + 3].find(self.Preserve2) != -1)
                    self.assertTrue(f[i + 4].find("{{{USER_STUFF_2") != -1)
                if not user3 and "{{{USER_STUFF_3" in line:
                    user3 = True
                    self.assertTrue(f[i + 1].find(self.Preserve3) != -1)
                    self.assertTrue(f[i + 2].find(self.Preserve3) != -1)
                    self.assertTrue(f[i + 3].find(self.Preserve3) != -1)
                    self.assertTrue(f[i + 4].find("{{{USER_STUFF_3") != -1)

    def checkPreservative(self, jam):
        self.assertIn(self.HeaderFileName, jam.preserved_tags_per_file, "Missing h")
        self.assertIn("{{{USER_STUFF_1", jam.preserved_tags_per_file[self.HeaderFileName], "Missing tag .h(1)")
        self.assertIn("{{{USER_STUFF_2", jam.preserved_tags_per_file[self.HeaderFileName], "Missing tag .h(2)")
        self.assertIn("{{{USER_STUFF_3", jam.preserved_tags_per_file[self.HeaderFileName], "Missing tag .h(3)")
        self.assertEqual(len(jam.preserved_tags_per_file[self.HeaderFileName]["{{{USER_STUFF_1"]), 3, "Wrong size of preserved data .h(1)")
        self.assertEqual(len(jam.preserved_tags_per_file[self.HeaderFileName]["{{{USER_STUFF_2"]), 3, "Wrong size of preserved data .h(2)")
        self.assertEqual(len(jam.preserved_tags_per_file[self.HeaderFileName]["{{{USER_STUFF_3"]), 3, "Wrong size of preserved data .h(3)")
        for line in jam.preserved_tags_per_file[self.HeaderFileName]["{{{USER_STUFF_1"]:
            self.assertTrue(line.find(self.Preserve1) != -1)
        for line in jam.preserved_tags_per_file[self.HeaderFileName]["{{{USER_STUFF_2"]:
            self.assertTrue(line.find(self.Preserve2) != -1)
        for line in jam.preserved_tags_per_file[self.HeaderFileName]["{{{USER_STUFF_3"]:
            self.assertTrue(line.find(self.Preserve3) != -1)

        self.assertIn(self.SourceFileName, jam.preserved_tags_per_file, "Missing cpp")
        self.assertIn("{{{USER_STUFF_1", jam.preserved_tags_per_file[self.SourceFileName], "Missing tag .cpp(1)")
        self.assertIn("{{{USER_STUFF_2", jam.preserved_tags_per_file[self.SourceFileName], "Missing tag .cpp(2)")
        self.assertIn("{{{USER_STUFF_3", jam.preserved_tags_per_file[self.SourceFileName], "Missing tag .cpp(3)")
        self.assertEqual(len(jam.preserved_tags_per_file[self.SourceFileName]["{{{USER_STUFF_1"]), 3, "Wrong size of preserved data .cpp(1)")
        self.assertEqual(len(jam.preserved_tags_per_file[self.SourceFileName]["{{{USER_STUFF_2"]), 3, "Wrong size of preserved data .cpp(2)")
        self.assertEqual(len(jam.preserved_tags_per_file[self.SourceFileName]["{{{USER_STUFF_3"]), 3, "Wrong size of preserved data .cpp(3)")
        for line in jam.preserved_tags_per_file[self.SourceFileName]["{{{USER_STUFF_1"]:
            self.assertTrue(line.find(self.Preserve1) != -1)
        for line in jam.preserved_tags_per_file[self.SourceFileName]["{{{USER_STUFF_2"]:
            self.assertTrue(line.find(self.Preserve2) != -1)
        for line in jam.preserved_tags_per_file[self.SourceFileName]["{{{USER_STUFF_3"]:
            self.assertTrue(line.find(self.Preserve3) != -1)

    def test_UseCase_FreshTemplateFromFileOnDisk(self):
        jam = Preservative(self.TempFolder)
        newFiles = OrderedDict()
        newFiles[self.HeaderFileName] = ['bla bla', '{{{USER_STUFF_2', '{{{USER_STUFF_2', 'crap', 'crap', 'crap', '{{{USER_STUFF_3', '{{{USER_STUFF_3', '{{{USER_STUFF_1', '{{{USER_STUFF_1', 'dang', ]
        newFiles[self.SourceFileName] = ['{{{USER_STUFF_3', '{{{USER_STUFF_3', 'bla bla', 'crap', 'crap', 'crap', '{{{USER_STUFF_1', '{{{USER_STUFF_1', '{{{USER_STUFF_2', '{{{USER_STUFF_2', 'dang', ]

        oldHeaderSize = len(newFiles[self.HeaderFileName])
        oldSourceSize = len(newFiles[self.SourceFileName])

        jam.Emplace(newFiles)

        # for each set of tags, the same text should be replaced 3 time...
        self.assertTrue((oldHeaderSize + 9) == len(newFiles[self.HeaderFileName]), "Expected at 9 more lines in " + self.HeaderFileName)
        self.assertTrue((oldSourceSize + 9) == len(newFiles[self.SourceFileName]), "Expected at 9 more lines in " + self.SourceFileName)

        # Check that the preserved text matches what is expected...
        self.checkPreservative(jam)
        # Check that the 'output' has the same expected things...
        self.checkOutput(newFiles)

    def test_UseCase_CloneFromOneFileToAnother(self):
        jam = Preservative(self.TempFolder)
        newFiles = OrderedDict()
        newFiles[self.HeaderFileName] = ['bla bla', '{{{USER_STUFF_2', "thisshouldgo2" ,'{{{USER_STUFF_2', 'crap', 'crap', 'crap', '{{{USER_STUFF_3', "thisshouldgo3" ,'{{{USER_STUFF_3', '{{{USER_STUFF_1', "thisshouldgo" , '{{{USER_STUFF_1', 'dang', ]
        newFiles[self.HeaderFileName].extend(["fish", "{{{USER_NOT_IN_OTHER", "BUTTHISSHOULDREMAIN", "{{{USER_NOT_IN_OTHER", "paste"])
        newFiles[self.SourceFileName] = ['{{{USER_STUFF_3', "thisshouldgo3" , '{{{USER_STUFF_3', 'bla bla', 'crap', 'crap', 'crap', '{{{USER_STUFF_1', "thisshouldgo" , '{{{USER_STUFF_1', '{{{USER_STUFF_2', "thisshouldgo2" , '{{{USER_STUFF_2', 'dang', ]
        newFiles[self.SourceFileName].extend(["fish", "{{{USER_NOT_IN_OTHER", "BUTTHISSHOULDREMAIN", "{{{USER_NOT_IN_OTHER", "paste"])

        oldHeaderSize = len(newFiles[self.HeaderFileName])
        oldSourceSize = len(newFiles[self.SourceFileName])

        jam.Emplace(newFiles, True)

        # for each set of tags, the same text should be replaced 3 time...but we replaced 1 line per tag
        self.assertTrue((oldHeaderSize + 9 - 3) == len(newFiles[self.HeaderFileName]), "Expected at 6 more lines in " + self.HeaderFileName)
        self.assertTrue((oldSourceSize + 9 - 3) == len(newFiles[self.SourceFileName]), "Expected at 6 more lines in " + self.SourceFileName)

        # Check that the preserved text matches what is expected...
        self.checkPreservative(jam)
        # Check that the unique tags are not in the preservative
        self.assertNotIn("{{{USER_NOT_IN_OTHER", jam.preserved_tags_per_file[self.HeaderFileName], "Unexpected tag .h(4)")
        self.assertNotIn("{{{USER_NOT_IN_OTHER", jam.preserved_tags_per_file[self.SourceFileName], "Unexpected tag .cpp(4)")
        # Check that the 'output' has the same expected things...
        self.checkOutput(newFiles)
        # Check that the things that were not in the source are still in the destination (and preserved)
        for name, f in newFiles.items():
            notClone = False
            self.assertIn("{{{USER_NOT_IN_OTHER", f, "Missing tag (unique) in output")
            for i in range(len(f)):
                line = f[i]
                if not notClone and "{{{USER_NOT_IN_OTHER" in line:
                    notClone = True
                    self.assertTrue(f[i + 1].find("BUTTHISSHOULDREMAIN") != -1)
                    self.assertTrue(f[i + 2].find("{{{USER_NOT_IN_OTHER") != -1)

if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python3
try:
    from . import protogen, smgen, umlgen, coggen, vppfs, LanguageCPP, LanguageCsharp, LanguagePython, cgen
except:
    import protogen, smgen, umlgen, coggen, vppfs, LanguageCPP, LanguageCsharp, LanguagePython, cgen

import os

''' Generate Entry function for Protocols. 
'''


#def Protocol(output_dir, pythoninterfacegeneratorfilename, namespacename, classname, declspec="", author = "", group="", brief="", template_dir="") -> list:
#    return protogen.Generate(output_dir, pythoninterfacegeneratorfilename, namespacename, classname, declspec, author, group, brief, template_dir)

def Protocol(outputdir, eventsinterface, namespacenname, classname, dclspc="", author="", group="", brief="", templatedir="", __internal="", __copy_other_files=True) -> list:
    if not templatedir.strip():
        templatedir = os.path.join(os.path.join(os.path.abspath(os.path.dirname(__file__)), "protocol_templates"), "CPP")
    res =  StateMachine(outputdir, [], eventsinterface, namespacenname, classname, dclspc, author, group, brief, templatedir, __internal, False)
    if __copy_other_files:
        protogen.CopyFrameworkFiles_CPP(outputdir)
    return res




''' Generate Entry function for State Machines
'''


def StateMachine(outputdir, transition_table, eventsinterface, namespacenname, statemachinenameprefix, dclspc="", author="", group="", brief="", templatedir="", __internal="", __copy_other_files=True) -> list:
    if not templatedir.strip():
        templatedir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "statemachine_templates_embedded_arm")

    if not os.path.isdir(templatedir) and not os.path.isfile(templatedir):
        print("Error : dir '" + templatedir + "' does not exist. Aborting.")
        return

    language = LanguageCPP.LanguageCPP()
    smgenerator = smgen.CStateMachineGenerator(templatedir, outputdir, eventsinterface, language, author, group, brief)
    if not __internal:
        smgenerator.vpp_filename = "Transition Table"
    else:
        smgenerator.vpp_filename = __internal
    return smgenerator.Generate(transition_table, namespacenname, statemachinenameprefix, dclspc, __copy_other_files)


def StateMachine_CSHARP(outputdir, transition_table, eventsinterface, namespacenname, statemachinenameprefix, dclspc="", author="", group="", brief="", templatedir="", __internal="", __copy_other_files=True) -> list:
    if not templatedir.strip():
        templatedir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "statemachine_templates_cs_winlinmac")

    if not os.path.isdir(templatedir) and not os.path.isfile(templatedir):
        print("Error : dir '" + templatedir + "' does not exist. Aborting.")
        return []

    language = LanguageCsharp.LanguageCsharp()
    smgenerator = smgen.CStateMachineGenerator(templatedir, outputdir, eventsinterface, language, author, group, brief)
    if not __internal:
        smgenerator.vpp_filename = "Transition Table"
    else:
        smgenerator.vpp_filename = __internal
    return smgenerator.Generate(transition_table, namespacenname, statemachinenameprefix, dclspc, __copy_other_files)


def StateMachine_PYTHON(outputdir, transition_table, eventsinterface, namespacenname, statemachinenameprefix, dclspc="", author="", group="", brief="", templatedir="", __internal="", __copy_other_files=True) -> list:
    if not templatedir.strip():
        templatedir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "statemachine_templates_py")

    if not os.path.isdir(templatedir) and not os.path.isfile(templatedir):
        print("Error : dir '" + templatedir + "' does not exist. Aborting.")
        return []

    language = LanguagePython.LanguagePython()
    smgenerator = smgen.CStateMachineGenerator(templatedir, outputdir, eventsinterface, language, author, group, brief)
    if not __internal:
        smgenerator.vpp_filename = "Transition Table"
    else:
        smgenerator.vpp_filename = __internal
    return smgenerator.Generate(transition_table, namespacenname, statemachinenameprefix, dclspc, __copy_other_files)


def StateMachineFromModel(outputdir, vp_project_path, vp_statemachinename, eventsinterface, namespacenname, statemachinenameprefix, dclspc="", author="", group="", brief="", templatedir="", __copy_other_files=True) -> list:
    transition_table = vppfs.ExtractTransitionTable(vp_statemachinename, vp_project_path)
    return StateMachine(outputdir, transition_table, eventsinterface, namespacenname, statemachinenameprefix, dclspc, author, group, brief, templatedir, os.path.basename(vp_project_path),__copy_other_files)


''' Generate Entry function for Class Diagrams
'''


def UML(outputdir, vp_project_path, vp_classdiagramname, dclspc="", author="", group="", brief="", namespace_to_folders=False, templatefiledir="") -> list:
    language = LanguageCPP.LanguageCPP()
    return umlgen.Generate(vp_project_path, vp_classdiagramname, outputdir, language, author, group, brief, namespace_to_folders, dclspc, templatefiledir)

def UML_CSHARP(outputdir, vp_project_path, vp_classdiagramname, dclspc="", author="", group="", brief="", namespace_to_folders=False, templatefiledir="") -> list:
    language = LanguageCsharp.LanguageCsharp()
    return umlgen.Generate(vp_project_path, vp_classdiagramname, outputdir, language, author, group, brief, namespace_to_folders, dclspc, templatefiledir)


''' Generate Entry function for template files using COG
'''


def Cogify(output_dir, pythonfile, cog_template_FILEorDIRECTORY, namespacename, classname, author="", group="", brief="", dclspec="") -> list:
    if os.path.isfile(cog_template_FILEorDIRECTORY):
        return coggen.GenerateFile(output_dir, pythonfile, cog_template_FILEorDIRECTORY, author, namespacename, classname, group, brief, dclspec)
    else:
        return coggen.GenerateDirectory(output_dir, pythonfile, cog_template_FILEorDIRECTORY, author, namespacename, classname, group, brief, dclspec)


''' File preservation sync utility
'''
def FileSync(from_file, to_file) -> None:
    cgen.FilePreservationSyncUtil(from_file, to_file)


import unittest

from kojen.LanguagePython import LanguagePython
from kojen.cgen import *

class TestFeatures(unittest.TestCase):

    workingfolder = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test")

    @classmethod
    def setUpClass(cls):
        os.makedirs(cls.workingfolder, exist_ok=True)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.workingfolder)

    @classmethod
    def create_template_file(cls, list_of_lines):
        assert os.path.isdir(cls.workingfolder)
        with open(os.path.join(cls.workingfolder, "file.t"), 'w') as f:
            for l in list_of_lines:
                f.write("%s\n" % l)

    @classmethod
    def get_test_gen(cls):
        TestFeatures.create_template_file(["Dummy"])
        return CGenerator(TestFeatures.workingfolder, TestFeatures.workingfolder, LanguagePython())

    def test_has_TAG(self):
        a = "XXX::blabla<<<something::1>>>"
        b = "XXX::blabla<<something>>"
        self.assertTrue(hasTag(a), "Failed to pick up tag")
        self.assertFalse(hasTag(b), "Failed to ignore tag")

    def test_has_specific_TAG(self):
        a = "XXX::blabla<<<something::1>>>"
        b = "XXX::blabla<<something>>"
        self.assertTrue(hasSpecificTag(a, "<<<something>>>"), "Failed to pick up specific tag")
        self.assertFalse(hasSpecificTag(b, "<<<something>>>"), "Failed to ignore specific tag")
        self.assertFalse(hasSpecificTag(a, "<<<nothing>>>"), "Failed to ignore specific tag")
        self.assertFalse(hasSpecificTag(b, "<<<nothing>>>"), "Failed to ignore specific tag")


    def test_TAG_has_default(self):
        a = "XXX::blabla<<<something::1>>>"
        b = "XXX::blabla<<<something>>>"
        self.assertTrue(hasDefault(a), "Failed to pick up default")
        self.assertFalse(hasDefault(b), "Failed to pick up default")

    def test_extract_default_and_TAG(self):
        a = "XXX::blabla<<<something::1>>>"
        b = "XXX::blabla<<<something>>>"
        c = "XXX::blabla<<something>>"
        d = "XXX::blabla<<<something::this::and::this>>>"
        res_a = extractDefaultAndTag(a)
        res_b = extractDefaultAndTag(b)
        res_c = extractDefaultAndTag(c)
        res_d = extractDefaultAndTag(d)
        self.assertEqual(res_a[0],"<<<something::1>>>", "Wrong tag")
        self.assertEqual(res_a[1],"1", "Wrong default")
        self.assertEqual(res_b[0], "<<<something>>>", "Wrong tag")
        self.assertEqual(res_b[1], "", "Wrong default")
        self.assertEqual(res_c[0], "", "Wrong tag")
        self.assertEqual(res_c[1], "", "Wrong default")
        self.assertEqual(res_d[0], "<<<something::this::and::this>>>", "Wrong tag")
        self.assertEqual(res_d[1], "this::and::this", "Wrong default")

    def test_extract_default_and_TAG_multiple(self):
        a = "XXX::blabla<<<something::1>>> !@#!@$ <<<else::2>>>"
        res_a = extractDefaultAndTagNamed(a, "something")
        res_b = extractDefaultAndTagNamed(a, "else")
        self.assertEqual(res_a[0], "<<<something::1>>>", "Wrong tag")
        self.assertEqual(res_a[1], "1", "Wrong default")
        self.assertEqual(res_b[0], "<<<else::2>>>", "Wrong tag")
        self.assertEqual(res_b[1], "2", "Wrong default")

    def test_extract_TAG_and_A_and_B(self):
        a = "blab @#$KLF!WEFJ <<<some::thing::here>>>"
        b = "blab @#$KLF!WEFJ <<<some::thing>>>"
        res_a = extractTagAndAandB(a)
        self.assertEqual(len(res_a), 3, "Wrong length")
        self.assertEqual(res_a[0], "<<<some::thing::here>>>", "Wrong tag")
        self.assertEqual(res_a[1], "thing", "Wrong A")
        self.assertEqual(res_a[2], "here", "Wrong B")
        res_b = extractTagAndAandB(b)
        self.assertEqual(len(res_b), 3, "Wrong length")
        self.assertEqual(res_b[0], "<<<some::thing>>>", "Wrong tag")
        self.assertEqual(res_b[1], "thing", "Wrong A")
        self.assertEqual(res_b[2], None, "Wrong B")

    def test_extract_TAG_and_A_and_B_multiple(self):
        a = "blab @#$KLF!WEFJ <<<some::thing>>> 123498123481234 <<<some::thing>>>"
        res_a = extractTagAndAandB(a)
        self.assertEqual(len(res_a), 3, "Wrong length")
        self.assertEqual(res_a[0], "<<<some::thing>>>", "Wrong tag")
        self.assertEqual(res_a[1], "thing", "Wrong A")
        self.assertEqual(res_a[2], None, "Wrong B")

    def test_remove_default(self):
        a = "XXX::blabla<<<something::1>>>"
        b = "XXX::blabla<<<something>>>"
        res_a = removeDefault(a)
        res_b = removeDefault(b)
        self.assertEqual(res_a,res_b, "Failed to remove default")
        self.assertEqual(res_b,res_b, "Failed to remove default")

    def test_alpha(self):
        for i in range(100):
            s = reset_alphabet()
            for j in range(26*2-1):
                s+=get_next_alphabet()
            self.assertEqual(s,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")


    def test_getWhitespace(self):
        a = "  <<<"
        b = "           <<<"
        res_a = getWhitespace(a)
        res_b = getWhitespace(b)
        self.assertEqual(res_a, "  ", "Failed to extract whitespace")
        self.assertEqual(res_b, "           ", "Failed to extract whitespace")


    def test_SingleExpander(self):
        all_lines = []
        all_lines.append("First")
        all_lines.append("      <<<MAGIC>>>")
        all_lines.append("Last")

        def basicExpansionFunction(output, whitespace):
            output.append(whitespace + "|||")
            output.append(whitespace + "---")
            output.append(whitespace + "***")

        se = SingleExpander("<<<MAGIC>>>")
        all_lines = se.Expand(all_lines, basicExpansionFunction)

        self.assertEqual(len(all_lines), 5, "unexpected input")
        self.assertEqual(all_lines[0], "First", "Unexpected output (1)")
        self.assertEqual(all_lines[1], "      |||", "Unexpected output (2)")
        self.assertEqual(all_lines[2], "      ---", "Unexpected output (3)")
        self.assertEqual(all_lines[3], "      ***", "Unexpected output (4)")
        self.assertEqual(all_lines[4], "Last", "Unexpected output (5)")

    def test_PairExpanderBasic(self):
        all_lines = []
        all_lines.append("First")
        all_lines.append("<<<BEGIN>>>")
        all_lines.append("---")
        all_lines.append(">>>")
        all_lines.append("<<<END>>>")
        all_lines.append("Last")

        def basicExpansionFunction(snippets, output):
            for s in snippets:
                output.append(s + "|||")
                output.append(s + "<<<")

        pe = PairExpander("<<<BEGIN>>>", "<<<END>>>")
        all_lines = pe.Expand(all_lines, basicExpansionFunction)

        self.assertEqual(len(all_lines), 6, "unexpected input")
        self.assertEqual(all_lines[0], "First", "Unexpected output (1)")
        self.assertEqual(all_lines[1], "---|||", "Unexpected output (2)")
        self.assertEqual(all_lines[2], "---<<<", "Unexpected output (3)")
        self.assertEqual(all_lines[3], ">>>|||", "Unexpected output (4)")
        self.assertEqual(all_lines[4], ">>><<<", "Unexpected output (5)")
        self.assertEqual(all_lines[5], "Last", "Unexpected output (6)")

    def test_PairExpanderExtraParam(self):
        all_lines = []
        all_lines.append("First")
        all_lines.append("<<<BEGIN::42>>>")
        all_lines.append("---")
        all_lines.append(">>>")
        all_lines.append("<<<END>>>")
        all_lines.append("Last")

        def extraParamExpansionFunction(snippets, output, extraParam):
            output.append("in : " + extraParam)
            for s in snippets:
                output.append(s + "|||")
                output.append(s + "<<<")

        pe = PairExpander("<<<BEGIN>>>", "<<<END>>>")
        all_lines = pe.Expand(all_lines, extraParamExpansionFunction)

        self.assertEqual(len(all_lines), 7, "unexpected input")
        self.assertEqual(all_lines[0], "First", "Unexpected output (1)")
        self.assertEqual(all_lines[1], "in : 42", "Unexpected output (2)")
        self.assertEqual(all_lines[2], "---|||", "Unexpected output (3)")
        self.assertEqual(all_lines[3], "---<<<", "Unexpected output (4)")
        self.assertEqual(all_lines[4], ">>>|||", "Unexpected output (5)")
        self.assertEqual(all_lines[5], ">>><<<", "Unexpected output (6)")
        self.assertEqual(all_lines[6], "Last", "Unexpected output (7)")

    def test_ForLoopExpanderCSV(self):

        def runTest(self, all_lines):
            test = TestFeatures.get_test_gen()
            output = PairExpander("<<<FOR_BEGIN>>>", "<<<FOR_END>>>").Expand(all_lines, test.innerexpand_for_loop)

            self.assertEqual(len(output), 13, "unexpected input")
            self.assertEqual(output[0], "First", "Unexpected output (1)")
            self.assertEqual(output[1], "fee", "Unexpected output (2)")
            self.assertEqual(output[2], "__fee__", "Unexpected output (3)")
            self.assertEqual(output[3], "A_fee = a", "Unexpected output (4)")
            self.assertEqual(output[4], "A_fee = 0", "Unexpected output (5)")
            self.assertEqual(output[5], "__fie__", "Unexpected output (6)")
            self.assertEqual(output[6], "A_fie = b", "Unexpected output (7)")
            self.assertEqual(output[7], "A_fie = 1", "Unexpected output (8)")
            self.assertEqual(output[8], "__foe__", "Unexpected output (9)")
            self.assertEqual(output[9], "A_foe = c", "Unexpected output (10)")
            self.assertEqual(output[10], "A_foe = 2", "Unexpected output (11)")
            self.assertEqual(output[11], "foe", "Unexpected output (12)")
            self.assertEqual(output[12], "Last", "Unexpected output (13)")

        all_lines = []
        all_lines.append("First")
        all_lines.append("<<<FOR_BEGIN::fee, fie, foe>>>")
        all_lines.append("<<<FIRST>>>")
        all_lines.append("__<<<EACH>>>__")
        all_lines.append("A_<<<EACH>>> = <<<ALPH>>>")
        all_lines.append("A_<<<EACH>>> = <<<NUM>>>")
        all_lines.append("<<<LAST>>>")
        all_lines.append("<<<FOR_END>>>")
        all_lines.append("Last")
        runTest(self, all_lines)
        # additional variants for the same output
        all_lines[1] = "<<<FOR_BEGIN::   fee, fie, foe >>>"
        runTest(self, all_lines)
        all_lines[1] = "<<<FOR_BEGIN:: ,  fee, fie, foe ,>>>"
        runTest(self, all_lines)
        all_lines[1] = "<<<FOR_BEGIN::   fee, fie, foe ,>>>"
        runTest(self, all_lines)
        all_lines[1] = "<<<FOR_BEGIN::  , fee, fie, foe     >>>"
        runTest(self, all_lines)

    def test_ForLoopExpanderNum(self):

        def runTest(self, all_lines):
            test = TestFeatures.get_test_gen()
            output = PairExpander("<<<FOR_BEGIN>>>", "<<<FOR_END>>>").Expand(all_lines, test.innerexpand_for_loop)

            self.assertEqual(len(output), 11, "unexpected input")
            self.assertEqual(output[0], "First", "Unexpected output (1)")
            self.assertEqual(output[1], "___0___", "Unexpected output (2)")
            self.assertEqual(output[2], "A__0_ = a", "Unexpected output (3)")
            self.assertEqual(output[3], "A__0_ = 0", "Unexpected output (4)")
            self.assertEqual(output[4], "___1___", "Unexpected output (5)")
            self.assertEqual(output[5], "A__1_ = b", "Unexpected output (6)")
            self.assertEqual(output[6], "A__1_ = 1", "Unexpected output (7)")
            self.assertEqual(output[7], "___2___", "Unexpected output (8)")
            self.assertEqual(output[8], "A__2_ = c", "Unexpected output (9)")
            self.assertEqual(output[9], "A__2_ = 2", "Unexpected output (10)")
            self.assertEqual(output[10], "Last", "Unexpected output (11)")

        all_lines = []
        all_lines.append("First")
        all_lines.append("<<<FOR_BEGIN::3>>>")
        all_lines.append("__<<<EACH>>>__")
        all_lines.append("A_<<<EACH>>> = <<<ALPH>>>")
        all_lines.append("A_<<<EACH>>> = <<<NUM>>>")
        all_lines.append("<<<FOR_END>>>")
        all_lines.append("Last")
        runTest(self, all_lines)


    ''' TODO : Testing
        - template extending and excluding.
        - user tags
          - only finds defaults after <<<
    '''

if __name__ == '__main__':
    unittest.main()


import unittest
import os
from kojen.vppfs import *

class TestDBFormat(unittest.TestCase):
    ''' Unit test. These guys might change the format of their DB.
        This hopes to make it easier to catch those changes and adapt to them quickly.
        This also is coded for the fixed number of diagrams in the model I do the dev with...if that changes, this code must change.
    '''
    def setUp(self):
        self.unit_test_model = os.path.join(os.path.abspath(os.path.dirname( __file__ )), "blob.xml")
        self.vppdiagrams = VPPDiagrams(self.unit_test_model)
        self.classdiagrams = self.vppdiagrams.GetClassDiagrams()
        self.statediagrams = self.vppdiagrams.GetStateDiagrams()

    def test_classDiagrams(self):
        expectedIDToName = {u'QcgxGqqFYEAQWAnB': u'ProtocolStack', u'sb74Kw6GAqAA7wjN': u'TestClassDiagram'}

        self.assertEqual(len(self.classdiagrams), 2									, "Expected 2 class diagrams in test project")

        for key,val in expectedIDToName.items():
            self.assertTrue((key in self.classdiagrams)								, "Class diagram '" + val + "' in TestModels.vpp test project ID changed.")
            self.assertEqual(self.classdiagrams[key], val						 	, "Class diagram '" + val + "' in TestModels.vpp test project NAME changed.")
            self.assertEqual(key, self.vppdiagrams.GetIDFromClassDiagramName(val)	, "Error in lookup class diagram " + val)

    def test_stateDiagrams(self):
        expectedIDToName = {u'4C0lKqqFYEAQWAg_': u'TestStateMachine'}

        self.assertEqual(len(self.statediagrams), 1									, "Expected 1 state diagram in test project")

        for key, val in expectedIDToName.items():
            self.assertTrue((key in self.statediagrams)								, "State diagram '" + val + "' in TestModels.vpp test project ID changed.")
            self.assertEqual(self.statediagrams[key], val							, "State diagram '" + val + "' in TestModels.vpp test project NAME changed.")
            self.assertEqual(key, self.vppdiagrams.GetIDFromStateDiagramName(val)	, "Error in lookup state diagram " + val)

    def test_stateDiagramTransitionTable(self):
        vppdiagramelements = VPPDiagramElements(self.unit_test_model)
        vppmodelelements = VPPModelElements(self.unit_test_model)
        elements = vppdiagramelements.GetDiagramElements(self.vppdiagrams.GetIDFromStateDiagramName('TestStateMachine'))

        self.assertEqual(len(elements),  8, "Expected 8 state diagram in elements in TestStateMachine in TestModels.vpp")

        # Create a state diagram with NAME, ID and CHILDREN ELEMENTs
        statediagram = StateDiagram('TestStateMachine', self.vppdiagrams.GetIDFromStateDiagramName('TestStateMachine'), elements, vppmodelelements)
        tt = statediagram.GetTransitionTable()

        self.assertEqual(len(tt), 3,		"Expected 3 rows in transition table from TestStateMachine in TestModels.vpp")

        self.assertTrue(tt[0][0] == 'StateRed' 		and tt[0][1] == 'EventButtonPressed' and tt[0][2] == 'StateOrange' 	and tt[0][3] == 'OnOrange' 	and tt[0][4] == 'GuardCanChangeToOrange', "Error with row 1 of transition table from TestStateMachine in TestModels.vpp")
        self.assertTrue(tt[1][0] == 'StateOrange' 	and tt[1][1] == 'EventButtonPressed' and tt[1][2] == 'StateGreen' 	and tt[1][3] == 'OnGreen' 	and tt[1][4] == 'GuardCanChangeToGreen', "Error with row 2 of transition table from TestStateMachine in TestModels.vpp")
        self.assertTrue(tt[2][0] == 'StateGreen' 	and tt[2][1] == 'EventButtonPressed' and tt[2][2] == 'StateRed' 	and tt[2][3] == 'OnRed' 	and tt[2][4] == 'GuardCanChangeToRed', "Error with row 3 of transition table from TestStateMachine in TestModels.vpp")

if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'eugene'

from collections import OrderedDict
import os, shutil
import datetime
import sys
import time
from pathlib import Path
try:
	from .preservative import *
except (ModuleNotFoundError, ImportError) as e:
	from preservative import *

import os

'''

    MIT License

    Copyright (c) 2015 Eugene Grobbelaar (email : koh.jaen@yahoo.de)

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

'''

''' This forms the base for some sorts of code generation.

    Step 1) Load template files to memory
    Step 2) Search and replace passed-in tags in memory (including filenames).
    
'''

__TAG_DATETIME__           = '<<<DATETIME>>>'
__TAG_PLATFORM__           = '<<<PLATFORM>>>'
__TAG_EXTENDS__            = '<<<EXTENDS>>>'
__TAG_EXCLUDE__            = '<<<EXCLUDE>>>'
__TAG_FOR_BEGIN__          = "<<<FOR_BEGIN>>>"
__TAG_EACH__               = "<<<EACH>>>"
__TAG_EACH_CAMELCAPS__     = "<<<each>>>"
__TAG_FIRST__              = "<<<FIRST>>>"
__TAG_LAST__               = "<<<LAST>>>"
__TAG_FOR_END__            = "<<<FOR_END>>>"
__TAG_ABC__                = '<<<ALPH>>>'
__TAG_123__                = '<<<NUM>>>'


'''------------------------------------------------------------------------------------------------------'''


class Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def warning(text) -> None:
    print(Colors.WARNING + "Warning : " + text + Colors.ENDC)


def info(text) -> None:
    print(Colors.OKGREEN + "Info : " + text + Colors.ENDC)


def error(text) -> None:
    print(Colors.FAIL + "Error : " + text + Colors.ENDC)


'''------------------------------------------------------------------------------------------------------'''


class CCodeModel:
    def __init__(self):
        self.filenames_to_lines = OrderedDict()

    def Merge(self, codemodel) -> None:
        """
        Will merge the input codemodel with this.

        @param codemodel: a CCodeModel object
        """
        self.filenames_to_lines.update(codemodel.filenames_to_lines)


'''------------------------------------------------------------------------------------------------------'''


class SingleExpander:
    def __init__(self, tag):
        self.tag = tag

    def Expand(self, all_lines, expansion_func, *args):
        all_lines_expanded = []
        for line in all_lines:
            if hasSpecificTag(line, self.tag):
                expansion_func(all_lines_expanded, getWhitespace(line), *args)
            else:
                all_lines_expanded.append(line)
        return all_lines_expanded


class PairExpander:
    def __init__(self, start_tag, end_tag):
        self.start_tag = start_tag
        self.end_tag = end_tag

    def Expand(self,  all_lines, expansion_func, *args):
        all_lines_expanded = []
        within_tags        = False
        snippet_to_expand  = []
        param              = None
        for line in all_lines:
            begin       = hasSpecificTag(line, self.start_tag)
            end         = hasSpecificTag(line, self.end_tag)
            within_tags = begin or within_tags
            if begin and hasDefault(line):
                param = extractDefaultAndTag(line)[1]
            if not within_tags and not end:
                all_lines_expanded.append(line)
            if within_tags and end:
                if param:
                    expansion_func(snippet_to_expand, all_lines_expanded, *args, param)
                else:
                    expansion_func(snippet_to_expand, all_lines_expanded, *args)
                snippet_to_expand = []
                within_tags = False
            if within_tags and not begin:
                snippet_to_expand.append(line)
        return all_lines_expanded


'''------------------------------------------------------------------------------------------------------'''
alpha = 97


def get_next_alphabet() -> str:
    global alpha
    alpha = alpha + 1
    if alpha == 123:
        alpha = 65
    if alpha == 91:
        alpha = 97
    return chr(alpha)


def reset_alphabet() -> str:
    global alpha
    alpha = 97
    return chr(alpha)


def even_space(a, nospaces=35) -> str:
    return a + (nospaces - len(a)) * " "


def camel_case(str) -> str:
    return str.title()


def camel_case_small(a) -> str:
    if a:
        return a[0].lower() + a[1:]
    return ""


def caps(a) -> str:
    return a.upper()


def hasTag(a):
    return '<<<' in a and '>>>' in a


def hasSpecificTag(a, tag):
    # allows for defaults : BEWARE also allows for partial matching so order is important in such a case.
    res = hasTag(a)
    if res:
        res = tag.replace("<<<", "").replace(">>>", "") in a
    return res


def hasDefault(a):
    return a.find("::", a.find("<<<")) >= 0

def extractDefaultAndTag(a):
    default = a[a.find("::", a.find("<<<")):a.rfind(">>>")].replace("::","", 1)
    tag = a[a.find("<<<"):a.rfind(">>>")+len(">>>")]
    return [tag, default]

def extractDefaultAndTagNamed(a, named):
    all = a.split(">>>")
    for b in all:
        if named in b:
            return extractDefaultAndTag(b + ">>>")
    raise Exception(named + " not found.")

def extractTagAndAandB(a):
    tag = a[a.find("<<<"):a.find(">>>") + len(">>>")]
    r = a[a.find("<<<") + len("<<<"):a.find(">>>")].split("::")
    A = None if len(r) < 2 else r[1]
    B = None if len(r) < 3 else r[2]
    return [tag, A, B]


def removeDefault(a):
    default = a[a.find("::", a.find("<<<")):a.rfind(">>>")]
    return a.replace(default, "")


def replaceDefault(a, b):
    default = a[a.find("::", a.find("<<<")):a.rfind(">>>")]
    default = default.strip("::")
    return a.replace(default, b)


def cleanTag(a):
    return a.replace("<<<","").replace(">>>","")


def getWhitespace(a):
    return a[0:a.find("<<<")]


'''------------------------------------------------------------------------------------------------------'''


class CGenerator:

    def __init__(self, inputfiledir, outputfiledir, language=None, author='Anonymous', group='', brief='',namespace_to_folders = False):
        self.input_template_file_dir = inputfiledir
        self.output_gen_file_dir = outputfiledir
        self.language = language
        self.author = author
        self.group = group
        self.brief = brief
        self.NAMESPACE_TO_GO_TO_OWN_FOLDER = namespace_to_folders
        # Does the input exist
        if not os.path.exists(inputfiledir):
            raise Exception("Directory '" + inputfiledir + "' does not exist.")
        else:
            files = None
            if os.path.isdir(inputfiledir):
                files = os.listdir(inputfiledir)
            if os.path.isfile(inputfiledir):
                files = inputfiledir
            # Is the input empty
            if not files:
                raise Exception("Directory '" + inputfiledir + "' is empty.")
            else:
                # Check the output dir
                if not os.path.exists(outputfiledir):
                    os.makedirs(outputfiledir)
                    info("Directory '" + outputfiledir + "' does not exist...created.")

    def generate_filenames_from_templates(self, file, dict_to_replace_filenames) -> str:
        for tag, desired_text in dict_to_replace_filenames.items():
            file = file.replace(tag, desired_text)
        return file

    ''' This will remove multiple newlines directly after each, leaving only 1'''
    def filter_multiple_newlines(self,list_of_lines_in_file):
        last = ''
        for i in range(len(list_of_lines_in_file)):
            was_filtered = False
            current = list_of_lines_in_file[i].replace(" ", "")
            if i > 0:
                if current == last and last == '\n':
                    list_of_lines_in_file[i] = ''
                    was_filtered = True
            if not was_filtered:
                last = current
        return list_of_lines_in_file


    def innerexpand_for_loop(self, to_expand, output, for_loop_param):

        def __process(csv_item_str, to_expand, output):
            alpha = reset_alphabet()
            cnt = 0
            first_processed = False
            last_processed = False
            first = None
            last = None
            items = csv_item_str.strip().lstrip(",").rstrip(",").split(',')
            to_add = []
            for i in items:
                for l in to_expand:
                    has_first = hasSpecificTag(l, __TAG_FIRST__)
                    has_last = hasSpecificTag(l, __TAG_LAST__)
                    if has_first and not first_processed:
                        first = l.replace(__TAG_FIRST__, items[0].strip())
                        first_processed = True
                    elif has_last and not last_processed:
                        last = l.replace(__TAG_LAST__, items[-1].strip())
                        last_processed = True
                    elif not has_first and not has_last:
                        to_add.append(l.replace(__TAG_EACH__, i.strip())
                                  .replace(__TAG_EACH_CAMELCAPS__, camel_case_small(i.strip()))
                                  .replace(__TAG_123__, str(cnt))
                                  .replace(__TAG_ABC__, alpha))
                cnt = cnt + 1
                alpha = get_next_alphabet()
            if first:
                to_add.insert(0, first)
            if last:
                to_add.append(last)
            output.extend(to_add)

        is_csv = for_loop_param.find(",") > -1
        is_numeric = for_loop_param.strip().isnumeric()
        if is_csv and not is_numeric:
            __process(for_loop_param, to_expand, output)
        elif not is_csv and is_numeric:
            number = int(for_loop_param.strip())
            new_for_loop_params = ""
            for i in range(number):
                new_for_loop_params += "_" + str(i) + "_" + ","
            __process(new_for_loop_params, to_expand, output)
        else:
            raise Exception("Unsupported FOR args.")


    def processExtends(self, path_to_parent_folder, inner_template_file, ignore_lines_with, ignore_lines_between):
        result = []

        templateFilePath = Path(inner_template_file)
        if not os.path.isabs(templateFilePath):
            templateFilePath = path_to_parent_folder / templateFilePath
        if not os.path.isfile(templateFilePath):
            warning("path '" + inner_template_file + "' does not exist. Ignoring.")
        else:
            with open(templateFilePath) as f:
                extended_lines = []
                isBetween = False
                isBetween_end = ''
                for l in f:
                    if not isBetween:
                        for start, end in ignore_lines_between:
                            if start in l:
                                isBetween = True
                                isBetween_end = end
                                break
                    if not isBetween:
                        if hasSpecificTag(l, __TAG_EXTENDS__): # nested extension
                            [unused, ext_rel_filepath] = extractDefaultAndTag(l)
                            if ext_rel_filepath:
                                nested_ext = self.processExtends(os.path.dirname(templateFilePath),ext_rel_filepath,ignore_lines_with, ignore_lines_between)
                                extended_lines.extend(nested_ext)
                        else:
                            extended_lines.append(l)
                    if isBetween:
                        if isBetween_end in l:
                            isBetween = False
                for l in extended_lines:
                    if not any(excl in l for excl in ignore_lines_with):
                        result.append(l)
        return result

    def processLine(self, dict_to_replace_lines, lines, line):
        for tag, desired_text in dict_to_replace_lines.items():
            desired_text = self.preserve_leading_tagwhitespace_in_multiline_searchandreplace(line, tag, desired_text)
            line = line.replace(tag, desired_text)
        # split multi-line-in-one-string to multi line. Code preservation does not work otherwise.
        if line.count('\n') > 1:
            lines_in_line = line.rstrip('\n').split('\n')
            for l in lines_in_line:
                lines.append(l + '\n')  # could do
        else:
            lines.append(line)

    def loadtemplates_firstfiltering_FILE(self, filepath, dict_to_replace_lines, dict_to_replace_filenames, filter_files_containing_in_name = "") -> CCodeModel:
        result = CCodeModel()
        if os.path.exists(filepath):
            file_without_path = os.path.basename(filepath)
            with open(filepath) as f:
                lines = []
                extended_filenames = []
                ignore_lines_with = []
                ignore_lines_between = []

                # Get all exclude tags...no matter what order they exist in the file.
                for line in f:
                    if hasSpecificTag(line, __TAG_EXCLUDE__):
                        [unused, exclude_line_with] = extractDefaultAndTag(line)
                        if exclude_line_with:
                            start_begin = exclude_line_with.split(",")
                            if len(start_begin) < 2:
                                ignore_lines_with.append(exclude_line_with)
                            elif len(start_begin) == 2:
                                ignore_lines_between.append(start_begin)
                            else:
                                warnings.warn("Tags '" + exclude_line_with + "' is not supported. Ignoring.")
                f.seek(0)
                for line in f:
                    if hasSpecificTag(line, __TAG_EXTENDS__):
                        [unused, ext_rel_filepath] = extractDefaultAndTag(line)
                        if ext_rel_filepath:
                            extension = self.processExtends(os.path.dirname(filepath), ext_rel_filepath, ignore_lines_with, ignore_lines_between)
                            for ex_l in extension:
                                # Replace the key:value pairs per line...
                                self.processLine(dict_to_replace_lines, lines, ex_l)
                            if extension:
                                # Replace the key:value pairs per filename...
                                self.processLine(dict_to_replace_filenames, extended_filenames, os.path.basename(ext_rel_filepath))
                    elif hasSpecificTag(line, __TAG_EXCLUDE__):
                        pass
                    else:
                        # Replace the key:value pairs per line...
                        self.processLine(dict_to_replace_lines, lines, line)

                # Replace any template names (extended) in the files by making them empty
                for i, l in enumerate(lines):
                    for ext_fn in extended_filenames:
                        if l.find(ext_fn) > -1:
                            lines[i] = ''

                # Replace the key:value pairs per filename...
                for tag, desired_text in dict_to_replace_filenames.items():
                    file_without_path = file_without_path.replace(tag, desired_text)
                # Remove multiple newlines
                lines = self.filter_multiple_newlines(lines)
                result.filenames_to_lines[file_without_path] = lines

                # would be great if user-tags and for loop processing could happen here
                # but that has problems with user-tags-without defaults
        return result

    def loadtemplates_firstfiltering(self, dict_to_replace_lines, dict_to_replace_filenames, filter_files_containing_in_name = ""):
        """
        Load Template and do 1st round of filtering. The filtering will replace the TAG

        @param dict_to_replace_lines: a dictionary of keys:values to replace per line
        @param dict_to_replace_filenames:  a dictionary of keys:values to replace per templatefilename. This includes extension.
        @param filter_files_containing_in_name:  fill process only files that contain this text in the name...or "" for all.
        @return: CCodeModel, a dictionary -> {filename,[lines]}
        """
        template_file_found = False
        result = CCodeModel()
        CWD = self.input_template_file_dir
        dict_to_replace_lines[__TAG_DATETIME__] = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
        dict_to_replace_lines[__TAG_PLATFORM__] = sys.platform + ' python ' + sys.version
        if os.path.isfile(self.input_template_file_dir):
            template_file_found = True
            cm = self.loadtemplates_firstfiltering_FILE(self.input_template_file_dir, dict_to_replace_lines, dict_to_replace_filenames, filter_files_containing_in_name)
            result.Merge(cm)
        else:
            for root, dirs, files in os.walk(CWD):
                for file in files:
                    if (file.lower().find(filter_files_containing_in_name.lower()) > -1 or not filter_files_containing_in_name.strip()) and not file.lower().find(".removed") > -1 :
                        template_file_found = True
                        cm = self.loadtemplates_firstfiltering_FILE(os.path.join(root, file), dict_to_replace_lines, dict_to_replace_filenames, filter_files_containing_in_name)
                        result.Merge(cm)

            if not template_file_found:
                raise Exception("Directory '" + self.input_template_file_dir + "' contains no templates.")

        return result

    def preserve_leading_tagwhitespace_in_multiline_searchandreplace(self, line, tag, desired_text):
        """
        For the case where the 'desired_text' that should replace the 'tag' in the 'line', if it is a multi-line
        replace, it will keep the leading spaces across all lines...otherwise simply returns the input desired_text
        @param line:
        @param tag:
        @param desired_text:
        @return:
        """
        if line.find(tag) != -1:
            desired_text_as_lines = desired_text.rstrip('\n').split('\n')
            if len(desired_text_as_lines) > 1:
                leading_spaces = (len(line) - len(line.lstrip(' '))) * " "
                desired_text = ""
                for d in desired_text_as_lines:
                    if not desired_text:
                        desired_text = d + "\n"
                    else:
                        desired_text = desired_text + leading_spaces + d + "\n"
                desired_text = desired_text.rstrip('\n')

        return desired_text

    def createoutput(self, filenames_to_lines) -> list:
        for f in filenames_to_lines:
            print("+++++++++ ", f)
            filename = os.path.join(self.output_gen_file_dir, f)
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, 'w') as writer:
                for line in filenames_to_lines[f]:
                    line = line.replace('\t',"    ") # Last filter! Convert tabs to 4 spaces...
                    writer.write(line)
        return list(filenames_to_lines.keys())


    def do_user_tags(self, codemodel, dict_key_vals):
        # get defaults : for loop processing introduces multiple uses of tags across multiple files.
        # first is the law!
        defaults_in_files_FOR = {}
        for fn, lines in codemodel.filenames_to_lines.items():
            for line in lines:
                has_tag = hasTag(line)
                has_for = hasSpecificTag(line, __TAG_FOR_BEGIN__)
                if has_tag and has_for:
                    fortaganddefault = extractDefaultAndTag(line)
                    if fortaganddefault[1].find("::") > -1:
                        hack = cleanTag(__TAG_FOR_BEGIN__)
                        hack = fortaganddefault[0].replace(hack+"::","")
                        if hasDefault(hack):
                            taganddefault = extractDefaultAndTag(hack)
                            key = cleanTag(removeDefault(taganddefault[0]))
                            if not taganddefault[1] in defaults_in_files_FOR:
                                defaults_in_files_FOR[key] = taganddefault[1]

        for fn, lines in codemodel.filenames_to_lines.items():
            new_lines = []

            # this should be called last, so at this point any tags should be user defined.
            for line in lines:
                has_tag      = hasTag(line)
                #has_user_tag = False
                has_for      = hasSpecificTag(line, __TAG_FOR_BEGIN__)
                #for k,v in dict_key_vals.items():
                #    if not has_user_tag:
                #        has_user_tag = hasSpecificTag(line, k)
                #        has_tag = False
                #        break
                if has_tag and not has_for:
                    taganddefault    = extractDefaultAndTag(line)
                    line             = removeDefault(line)
                    taganddefault[0] = removeDefault(taganddefault[0])
                    tagnoprepostfix  = taganddefault[0].replace('<<<','').replace('>>>','')
                    if tagnoprepostfix in dict_key_vals:
                        line = line.replace(taganddefault[0], str(dict_key_vals[tagnoprepostfix]))
                    elif taganddefault[1].strip():
                        line = line.replace(taganddefault[0], taganddefault[1])
                elif has_tag and has_for:
                    taganddefault = extractDefaultAndTag(line)
                    key = cleanTag(removeDefault("<<<" + taganddefault[1] + ">>>"))
                    if key in dict_key_vals:
                       line = replaceDefault(line, dict_key_vals[key])
                    elif key in defaults_in_files_FOR:
                        line = replaceDefault(line, defaults_in_files_FOR[key])
                    else:
                        line = "//POO"

                new_lines.append(line)
            # replace
            codemodel.filenames_to_lines[fn] = new_lines


    def do_for(self, codemodel):
        for fn, lines in codemodel.filenames_to_lines.items():
            new_lines = PairExpander(__TAG_FOR_BEGIN__, __TAG_FOR_END__).Expand(lines, self.innerexpand_for_loop)
            # replace
            codemodel.filenames_to_lines[fn] = new_lines

    '''Will use the base-class configured 'output directory' if no preserve directory is passed in. '''
    def preserve_usercode_in_files(self, codemodel, preserve_dir = ""):
        copy_filename_to_lines = codemodel.filenames_to_lines.copy() # prevent mutation whilst iteration.
        for filename_nopath in copy_filename_to_lines:
            file_to_preserve = ""
            if preserve_dir == "":
                file_to_preserve = os.path.join(self.output_gen_file_dir, filename_nopath)
            else:
                file_to_preserve = os.path.join(preserve_dir, filename_nopath)
            preservation = Preservative(file_to_preserve)
            preservation.Emplace(codemodel.filenames_to_lines)
'''------------------------------------------------------------------------------------------------------'''


def FileCopyUtil(dir_from, dir_to, list_of_filenames):
    """
    Will copy each file from list_of_filenames in dir_from to dir_to.
    Will create dir_to (even if its a tree) if it does not exist.

    @param dir_from: The directory from, where the list of files reside.
    @param dir_to: The directory the list of files should be copied to.
    @param list_of_filenames: The list [] of filenames to be copied.
    """
    try:
        os.makedirs(dir_to, exist_ok=True)
        for filename in list_of_filenames:
            try:
                shutil.copy(os.path.join(dir_from, filename), os.path.join(dir_to, filename))
            except OSError:
                warnings.warn("Copy of the file %s failed" % os.path.join(dir_from, filename))
    except OSError:
        warnings.warn("Creation of the directory %s failed" % dir_to)

def FilePreservationSyncUtil(file_from, file_to) -> None:
    """
    Will synchronize code in preservation tags from 'file_from' to 'file_to',
    if these tags exist in 'file_to'.

    Tags that exist exclusively in 'file_to' will remain untouched if they do not
    exist in 'file_from'.
    """
    if not os.path.isfile(file_from):
        error("File '" + file_from + "' does not exist. Aborting.")
        return
    if not os.path.isfile(file_to):
        error("File '" + file_to + "' does not exist. Aborting.")
        return

    print("*************************************")
    print("******* Sync (file)  ****************")
    print("*************************************")
    print(" From file    : " + file_from)
    print(" To file      : " + file_to)
    print(" Executing in : " + os.path.realpath(__file__))
    print("*************************************")

    bg = CGenerator(os.path.dirname(file_from), os.path.dirname(file_to))
    cm = bg.loadtemplates_firstfiltering_FILE(file_to, {}, {})
    p  = Preservative(file_from)
    p.preserved_tags_per_file[file_to]          = p.preserved_tags_per_file.pop(file_from)
    p.preserved_tags_per_file_WAS_USED[file_to] = p.preserved_tags_per_file_WAS_USED.pop(file_from)
    p.Emplace(cm.filenames_to_lines, True)
    bg.createoutput(cm.filenames_to_lines)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'eugene'

'''

    MIT License

    Copyright (c) 2015 Eugene Grobbelaar (email : koh.jaen@yahoo.de)

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

'''

__TAG_AUTHOR__              = '<<<AUTHOR>>>'
__TAG_NAMESPACE__           = '<<<NAMESPACE>>>'
__TAG_CLASS_NAME__          = '<<<CLASSNAME>>>'
__TAG_PyIFGen_NAME__        = '<<<PYIFGENNAME>>>'
__TAG_GROUP__               = '<<<GROUP>>>'
__TAG_BRIEF__               = '<<<BRIEF>>>'
__TAG_DECLSPEC_DLL_EXPORT__ = '<<<DLL_EXPORT>>>'

try:
    from .preservative import *
except (ModuleNotFoundError, ImportError) as e:
    from preservative import *

try:
    from .cgen import CGenerator, CCodeModel, FileCopyUtil
except (ModuleNotFoundError, ImportError) as e:
    from cgen import CGenerator, CCodeModel, FileCopyUtil

from cogapp import *
import shutil

class CCogCodeModel:
    def __init__(self):
        self.namespacename    = ""
        self.declspecdllexport = ""
        self.pythoninterfacegeneratorfilename = ""
        self.classname = ""
        self.group = ""
        self.brief = ""

class CCogGenerator(CGenerator):
    def __init__(self, inputfiledir, outputfiledir,  classname, language=None, author='Anonymous', group='', brief=''):
        CGenerator.__init__(self, inputfiledir, outputfiledir, language, author, group, brief)
        self.dict_to_replace_filenames = {}
        self.dict_to_replace_filenames["TEMPLATE_"] = classname

    def loadtemplates_firstfiltering(self, ccmodel):
        """
        See baseclass implementation. This just prepares the dictionary of things to replace
        for this type of codegeneration.

        @param smmodel:
        @return: cgen.CCodeModel, a dictionary -> {filename,[lines]}
        """

        dict_to_replace_lines = {}
        dict_to_replace_lines[__TAG_CLASS_NAME__] = ccmodel.classname
        dict_to_replace_lines[__TAG_PyIFGen_NAME__] = ccmodel.pythoninterfacegeneratorfilename.replace('.py','')  # hack : for tcpgen simple templates,
        dict_to_replace_lines[__TAG_NAMESPACE__] = ccmodel.namespacename
        dict_to_replace_lines[__TAG_AUTHOR__] = self.author
        dict_to_replace_lines[__TAG_DECLSPEC_DLL_EXPORT__] = ccmodel.declspecdllexport
        dict_to_replace_lines[__TAG_GROUP__] = ccmodel.group
        dict_to_replace_lines[__TAG_BRIEF__] = ccmodel.brief

        return CGenerator.loadtemplates_firstfiltering(self, dict_to_replace_lines, self.dict_to_replace_filenames)

    def generate_filenames_from_templates(self, file) -> str:
        return CGenerator.generate_filenames_from_templates(self,file, self.dict_to_replace_filenames)

    def Generate(self, pythoninterfacegeneratorfilename, namespacenname, classname, group, brief, preserve_dir="",dclspc="") -> list:
        sm = CCogCodeModel()
        sm.pythoninterfacegeneratorfilename = pythoninterfacegeneratorfilename
        sm.namespacename = namespacenname
        sm.declspecdllexport = dclspc
        sm.classname = classname
        sm.group = group
        sm.brief = brief

        self.cm = self.loadtemplates_firstfiltering(sm)

        # Preserve user code.
        self.preserve_usercode_in_files(self.cm,preserve_dir)

        # Write output to file, and return filenames.
        return self.createoutput(self.cm.filenames_to_lines)

def PreCogCopyFileAndInsertPythonImport(file_from, file_to, python_import):
    # Need to put the correct python file for cog to import!
    with open(file_from, "rt") as fin:
        with open(file_to, "wt") as fout:
            for line in fin:
                fout.write(line.replace(__TAG_PyIFGen_NAME__, python_import))

def GenerateFile(output_dir, pythonfile, cog_template_file, author, namespacename, classname, group, brief, dclspc="") -> list:
    cog_template_file = cog_template_file.strip()
    if os.path.isdir(cog_template_file):
        return GenerateDirectory(output_dir, pythonfile, cog_template_file, author, namespacename, classname, group, brief, dclspc)
    if not os.path.isfile(cog_template_file):
        print("Error : file '" + cog_template_file + "' does not exist. Aborting.")
        return []

    print("*************************************")
    print("******* CogGen (file)****************")
    print("*************************************")
    print(" Template File: " + cog_template_file)
    print(" Output Dir   : " + output_dir)
    print(" Python File  : " + pythonfile)
    print(" Executing in : " + os.path.realpath(__file__))
    print("*************************************")

    cog_template_dir = os.path.dirname(cog_template_file)
    cogged_template_dir = os.path.join(cog_template_dir, "cogged")
    if not os.path.exists(cogged_template_dir):
        os.makedirs(cogged_template_dir)
    # copy the python file to the cogged directory, pre-cog...
    tmp_python_file = os.path.join(cogged_template_dir, os.path.basename(pythonfile))
    shutil.copy(pythonfile, tmp_python_file)
    # copy the cog template to the cogged directory, pre-cog...and rename it as such. This is the first pass.
    file_to_cog = os.path.join(cogged_template_dir,os.path.basename(cog_template_file))
    PreCogCopyFileAndInsertPythonImport(cog_template_file, file_to_cog + ".PreCog", os.path.splitext(os.path.basename(tmp_python_file))[0])
    # Create a generator
    gen = CCogGenerator(cogged_template_dir, output_dir, classname, None, author)
    with open(os.path.join(cogged_template_dir, "files.txt"), 'w') as writer:
        writer.write('"' + file_to_cog + ".PreCog" + '" -o "' + file_to_cog + '"\r\n')
    # Invoke Cog on intermediate dir
    c = Cog()
    c.callableMain(['cog', '-d', '@' + os.path.join(cogged_template_dir, "files.txt")])
    # clean up cog file and python file
    os.remove(os.path.join(cogged_template_dir, "files.txt"))
    os.remove(tmp_python_file)
    os.remove(file_to_cog + ".PreCog")
    # Do rest...i.e. replace some stuff...preserve...
    res = gen.Generate(pythonfile, namespacename, classname, group, brief, output_dir, dclspc)
    # Finally : delete the cogged folder
    shutil.rmtree(cogged_template_dir)
    return res

def GenerateDirectory(output_dir, pythonfile, cog_template_dir, author, namespacename, classname, group, brief, dclspc="") -> list:
    cog_template_dir = cog_template_dir.strip()
    if os.path.isfile(cog_template_dir):
        return GenerateFile(output_dir, pythonfile, cog_template_dir, author, namespacename, classname, group, brief, dclspc)
    if not os.path.isdir(cog_template_dir):
        print("Error : dir '" + cog_template_dir + "' does not exist. Aborting.")
        return []

    print("*************************************")
    print("******* CogGen (directory) **********")
    print("*************************************")
    print(" Template Dir : " + cog_template_dir)
    print(" Output Dir   : " + output_dir)
    print(" Python File  : " + pythonfile)
    print(" Executing in : " + os.path.realpath(__file__))
    print("*************************************")

    cogged_template_dir = os.path.join(cog_template_dir, "cogged")
    if not os.path.exists(cogged_template_dir):
        os.makedirs(cogged_template_dir)
    # copy the python file to the cogged directory, pre-cog...
    tmp_python_file = os.path.join(cogged_template_dir, os.path.basename(pythonfile))
    shutil.copy(pythonfile, tmp_python_file)
    # copy the cog templates to the cogged directory, pre-cog...and rename it as such. This is the first pass.
    files_to_cog = []
    for root, dirs, files in os.walk(cog_template_dir):
        if  "cogged" in dirs:
            dirs.remove("cogged")
        for file in files:
            file_to_cog = os.path.join(cogged_template_dir,os.path.basename(file))
            PreCogCopyFileAndInsertPythonImport(os.path.join(cog_template_dir,file),file_to_cog+".PreCog",os.path.splitext(os.path.basename(tmp_python_file))[0])
            files_to_cog.append(file_to_cog)

    # Create a generator
    gen = CCogGenerator(cogged_template_dir, output_dir, classname, None, author)
    # Create Cog Input
    with open(os.path.join(cogged_template_dir, "files.txt"), 'w') as writer:
        for f in files_to_cog:
            writer.write('"' + f + ".PreCog" + '" -o "' + f + '"\r\n')
    # Invoke Cog on intermediate dir
    c = Cog()
    c.callableMain(['cog', '-d', '@' + os.path.join(cogged_template_dir, "files.txt")])
    # clean up cog file and python file
    os.remove(os.path.join(cogged_template_dir, "files.txt"))
    os.remove(tmp_python_file)
    for f in files_to_cog:
        os.remove(f + ".PreCog")
    # Do rest...i.e. replace some stuff...preserve...
    res = gen.Generate(pythonfile, namespacename, classname, group, brief, output_dir, dclspc)
    # Finally : delete the cogged folder
    shutil.rmtree(cogged_template_dir)
    return res


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'eugene'

'''

    MIT License

    Copyright (c) 2015 Eugene Grobbelaar (email : koh.jaen@yahoo.de)

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

'''
import datetime
import sys
import time
from string import Template

try:
    from kojentypes import *
except (ModuleNotFoundError, ImportError) as e:
    from .kojentypes import *

try:
    from vppclassdiagram import *
except (ModuleNotFoundError, ImportError) as e:
    from .vppclassdiagram import *

try:
    from Language import *
except (ModuleNotFoundError, ImportError) as e:
    from .Language import *


class LanguageCPP(Language):

    '''USED'''
    def ByteStreamTypeSharedPtr(self):
        return self.SharedPtrToType(self.ByteStreamType())

    '''USED'''
    def ByteStreamTypeRawPtr(self):
        return self.RawPtrToType("uint8")

    '''USED'''
    def ByteStreamType(self):
        return 'std::vector<uint8>'

    '''USED'''
    def TypedefSharedPtrToType(self, typename):
        return 'typedef ' + self.SharedPtrToType(typename) + " " + self.PtrToTypeName(typename)

    '''USED'''
    def TypedefRawPtrToType(self, typename):
        return 'typedef ' + self.RawPtrToType(typename) + " " + self.PtrToTypeName(typename)

    def PtrToTypeName(self, typename):
        return typename + '_ptr'

    '''USED'''
    # Convenience
    def SharedPtrToType(self, typename):
        return 'justaplain_shared_ptr<' + typename + '>'

    def RawPtrToType(self, typename):
        return typename + '*'

    # parameters need to be a list of (type, name).
    def ParameterString(self, parameters=None) -> str:
        if parameters is None:
            parameters = []

        if str(type(parameters)) == "<type 'list'>" or str(type(parameters)) == "<class 'list'>":  # Python2 gives the first, python3 the second!
            parameter_string = ''
            size = len(parameters)
            cnt = 0
            for param in parameters:
                if HasDefault(param):
                    parameter_string += param[0] + ' ' + param[1] + "=" + param[2]
                else:
                    parameter_string += param[0] + ' ' + param[1]
                if cnt != (size - 1):
                    parameter_string += ', '
                cnt += 1
            return parameter_string

        raise Exception("Please use OrderedDict when passing parameters into 'ParameterString'")
    '''USED'''
    def DeclareFunction(self, returntype, classname, functionname, is_impl, parameters=None, virtual=False, is_static=False, is_const=False) -> str:
        # def DeclareFunction(self, returntype, classname, functionname, is_impl, parameters=[], virtual=False):
        # https://florimond.dev/blog/articles/2018/08/python-mutable-defaults-are-the-source-of-all-evil/
        if parameters is None:
            parameters = []

        if str(type(parameters)) == "<type 'list'>" or str(type(parameters)) == "<class 'list'>":  # Python2 gives the first, python3 the second!
            parameter_string = self.ParameterString(parameters)

            if is_impl:
                if classname.replace(' ', '') != '':
                    return returntype + ' ' + classname + '::' + functionname + '(' + parameter_string + ')' + (" const" if is_const else "")
                return returntype + ' ' + functionname + '(' + parameter_string + ')' + (" const" if is_const else "")

            if virtual and not is_static:  # don't allow static.
                return 'virtual ' + returntype + ' ' + functionname + '(' + parameter_string + ')' + (" const" if is_const else "")

            return ("static " if is_static else "") + returntype + ' ' + functionname + '(' + parameter_string + ')' + (" const" if is_const else "")

        raise Exception("Please use list when passing parameters into 'DeclareFunction'")

    '''USED'''
    def GetFactoryCreateParams(self, struct, interface, with_defaults=False) -> list:
        def _processDefaults(default) -> str:
            res = ""
            if str(type(default))=="<class 'list'>":
                res += "{"
                for i in default:
                    if HasDefault(i):
                        res += _processDefaults(i[2]) + ","
                    else:
                        res += "{},"
                res = res.rstrip(",")
                res += "}"
            else:
                res += default
            return res

        structmembers = struct.Decompose()
        factoryparams = []
        for mem in structmembers:
            isMessage = interface.IsMessageStruct(mem[0])
            isStruct = struct.IsStruct(mem[1])
            if not interface.IsProtocolStruct(mem[0]):
                ref = " const&" if isStruct or isMessage else ""
                if (mem[0] in interface):
                    factoryparams.append(((self.SharedPtrToType(mem[0]) if isMessage else mem[0]) + ref,
                                          (mem[1] + "=" + _processDefaults(mem[2])) if (with_defaults and HasDefault(mem)) else mem[1]))
                else:
                    if with_defaults:
                        if HasDefault(mem):
                            factoryparams.append((mem[0] + ref, mem[1], _processDefaults(mem[2])))
                        else:
                            factoryparams.append((mem[0] + ref, mem[1], "{}"))
                    else:
                        factoryparams.append((mem[0] + ref, mem[1]))
        return factoryparams

    '''USED
    Declares the guts of a struct declaration
    '''
    def DeclareStructMembers(self, struct, interface, whitespace, attr_packed=True) -> list:
        arrayName = []
        structmembers = struct.Decompose()
        result = []
        for mem in structmembers:
            if (mem[0] in interface):
                result.append(whitespace + self.InstantiateType(mem[0], mem[1]) + ";")
            else:
                result.append(whitespace + self.InstantiateType(mem[0], mem[1],
                                                                "" if (not HasDefault(mem) or attr_packed) else mem[2],
                                                                attr_packed) + ";")
        return result
    '''USED'''
    def InstantiateStructMembers(self, struct, interface, whitespace, instancename, accessor) -> list:
        structmembers = struct.Decompose()
        result = []
        for mem in structmembers:
            isStruct = interface.IsStruct(mem[1])
            isProtocol = interface.IsProtocolStruct(mem[0])
            instance_accessor = whitespace + instancename + accessor

            if not isProtocol or isStruct:
                result.append(instance_accessor + self.InstantiateType('', mem[1], mem[1]) + ";")
            elif isProtocol and not isStruct:
                s = Template("sizeof(${this}) - sizeof(${header})")
                # Aggregate initializer
                new = instance_accessor +  self.InstantiateType("", mem[1], "{" + struct[mem[1]].GetDefaultsAsString(s.substitute(this=struct.Name, header=mem[0])) + "};")
                result.append(new)
            else:
                print("WTF : InstantiateStructMembers")
        return result

    def InstantiateStructWithAggregateInitializer(self, struct, interface) -> str:
        structmembers = struct.Decompose()
        result = ""
        for mem in structmembers:
            isStruct = interface.IsStruct(mem[0])
            isProtocol = interface.IsProtocolStruct(mem[0])
            if not isProtocol or isStruct:
                result += mem[1] + ", "
            elif isProtocol:
                s = Template("sizeof(${this}) - sizeof(${header})")
                result += "{" + struct[mem[1]].GetDefaultsAsString(s.substitute(this=struct.Name, header=mem[0])) + "}, "
            else:
                raise Exception("InstantiateStructWithAggregateInitializer unhandled!")
        return "{" + result.rstrip(", ") + "}"

    def DefaultAggregateInitializer(self) -> str:
        return "{}"


    '''USED'''
    def InstantiatePtrToType(self, typename, instancename, typepointername=""):
        return (typepointername if typepointername else self.PtrToTypeName(typename)) + ' ' + instancename + '(new ' + typename + ')'

    '''USED'''
    def InstantiateArray(self, typename, instancename, noelements):
        return instancename + ' = new ' + typename + '['+noelements+']'

    '''USED

        Instantiate/declare a basic type, with optional initializer. Only declarations can use 'attribute packed' directives.
        use typename = '' to make it a known variable initialization
    '''
    def InstantiateType(self, typename, instancename, initialevalue='', is_attr_packed=False, is_static=False, is_const=False) -> str:
        # declaration...i.e. in class or struct. no initial value, can be packed
        if initialevalue.replace(' ', '') == '':
            decl = typename + ' ' + instancename
            if is_static:
                decl = 'static ' + decl
            if is_const:
                decl = 'const ' + decl
            if is_attr_packed:
                return self.AddAttributePackedToDecl(decl)
            return decl

        # local variable decl/instantiation or known variable instantiation
        if typename.replace(' ', '') == '':
            return instancename + ' = ' + initialevalue

        res = typename + ' ' + instancename + ' = ' + initialevalue
        if is_static:
            res = 'static ' + res
        if is_const:
            res = 'const ' + res
        return res

    # Braces
    '''USED'''
    def OpenBrace(self):
        return '{'

    '''USED'''
    def CloseBrace(self):
        return '}'

    # Connection type
    def ConnectionType(self):
        return 'IConnection'

    def ConnectionTypePtr(self):
        return 'IConnection_ptr'

    # Member access / dereference
    '''USED'''
    def Accessor(self, is_ptr):
        if is_ptr:
            return '->'
        return '.'

    # Bytestream access / dereference
    def BytestreamAccessor(self):
        return '->'

    # Packedness
    def AddAttributePackedToDecl(self, declaration):
        ws = ' '
        len_declaration = len(declaration)
        trailingspacecnt = 50 - len_declaration
        for i in range(trailingspacecnt):
            ws = ws + ' '
        result = declaration + ws + '__attribute__ ((packed))'
        return result

    def DeleteArray(self, instancename):
        return 'delete[] ' + instancename + ';\n'

    def DeletePtrToType(self, instancename):
        return 'delete ' + instancename + ';\n'

    def DeclareClass(self, classname, declspec=''):
        if declspec.replace(' ', '') != '':
            return 'class ' + declspec + ' ' + classname + '\n'
        return 'class ' + classname + '\n'

    def ForwareDeclareClass(self, classname):
        return 'CGEN_DECL_CLASS_PTR(' + classname + ');\n'

    def DeclareStruct(self, structname, declspec=''):
        if declspec.replace(' ', '') != '':
            return 'struct ' + declspec + ' ' + structname + '\n'
        return 'struct ' + structname + '\n'

    def DeclareEnum(self, enum, whitespace) -> str:
        #result = ""
        #if hasattr(enum, 'documentation'):
        #    result += self.FormatComment(enum.documentation)
        result = self.FormatComment(enum.documentation) + "\n"
        result += "enum class " + enum.Name + " : unsigned char\n"
        result += "{\n"
        for descriptionName, val in enum.items():
            result += whitespace + str(descriptionName) + " = " + str(val) + ",\n"
        result = result[:len(result)-2] + "\n"  # items from the beginning through end-1 (i.e. remove last character which is a ','
        result += "};\n"
        return result

    def DeclareHashDefine(self, name, val):
        return "#define " + name + " " + str(val)

    def DeclareNamespace(self, namespacename):
        return 'namespace ' + namespacename + '\n'

    def UsingNamespace(self, namespacename):
        return 'using namespace ' + namespacename + ';\n'
    '''USED'''
    def FormatComment(self, commenttext):
        if commenttext:
            return '// ' + commenttext
        return ""

    def FormatLongComment(self, commenttext):
        if commenttext:
            return '/**' + commenttext + '*/\n'
        return ""

    # The includes that each generated file needs to have from the provided mini framework
    def LanguageSpecificFrameWorkIncludes(self):
        return ['"../../allplatforms/basetypes.h"']

    def AddInclude(self, filename):  # for python we can simply replace .h with''
        return '#include ' + filename

    def PrintError(self, message):
        return 'printf( ' + message + ');'
    '''USED'''
    def PrintMessage(self, message):
        return 'printf(' + message + ');'

    def BooleanTrue(self):
        return 'true'

    def BooleanFalse(self):
        return 'false'

    '''USED'''
    def NullPtr(self):
        return 'nullptr'

    def PublicAccessSpecifier(self):
        return 'public:\n'

    def ProtectedAccessSpecifier(self):
        return 'protected:\n'

    def PrivateAccessSpecifier(self):
        return 'private:\n'

    def If(self, statement):
        return 'if('+statement+')'

    def ElseIf(self, statement):
        return 'else if('+statement+')'

    def Else(self):
        return 'else'

    def This(self):
        return 'this->'

    # ------------------------------ End

    def LicenseAgreement(self):
        product_name = "'KoJen'"
        result = []
        result.append("--------------------------------------------------------------------------------")
        result.append("")
        result.append('    ' + "This file is part of " + product_name + ".")
        result.append("")
        result.append('    ' + product_name + " is free software: you can redistribute it and/or modify")
        result.append('    ' + "it under the terms of the GNU General Public License as published by")
        result.append('    ' + "the Free Software Foundation, either version 3 of the License, or")
        result.append('    ' + "(at your option) any later version.")
        result.append("")
        result.append('    ' + product_name + " is distributed in the hope that it will be useful,")
        result.append('    ' + "but WITHOUT ANY WARRANTY; without even the implied warranty of")
        result.append('    ' + "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the")
        result.append('    ' + "GNU General Public License for more details.")
        result.append("")
        result.append('    ' + "You should have received a copy of the GNU General Public License")
        result.append('    ' + "along with " + product_name + ".  If not, see <http://www.gnu.org/licenses/>.")
        result.append('    ' + "For any queries please contact : koh.jaen@yahoo.de.")
        result.append("\n")
        result.append('        ' + "This file was generated using : " + sys.platform + ".")
        result.append('        ' + "This file was generated by a machine. Do not modify it by hand.")
        result.append("")
        result.append("--------------------------------------------------------------------------------")
        return result

    ''' Additions that were added for class diagram
    '''

    def GetFormatNestedNamespaceBegin(self, classObj):
        """
        Returns the C++ format string for a nested namespace declaration on one line...
        @param classObj: class information from vppclassdiagram.Class
        @return: string i.e. "namespace XFullStack { namespace XProtocol {"
        """
        # Nested NAMESPACE of the classobj is i.e. 'XFullStack::XProtocol'
        if isinstance(classObj, Class):
            result = _getFormatNestedNamespaceBegin(classObj.NAMESPACE)
            return result
        else:
            raise RuntimeError("Language Feature Not Implemented")

    def GetFormatNestedNamespaceEnd(self, classObj):
        """
         Returns the C++ format string for a nested namespace end-declaration on one line...

        @param classObj: class information from vppclassdiagram.Class
        @return: string i.e. for "namespace XFullStack { namespace XProtocol {" will return "} }"
        """
        # Nested NAMESPACE of the classobj is i.e. 'XFullStack::XProtocol'
        if isinstance(classObj, Class):
            result = _getFormatNestedNamespaceEnd(classObj.NAMESPACE)
            return result + " // end namespace " + classObj.NAMESPACE
        else:
            raise RuntimeError("classObj Not of type Class")

    def GetFormatClassInheritence(self, classObj, dictOfInheritance):
        """
        Returns the C++ format string for class inheritence based on objects
        returned by parsing the class diagram in e.g. our favourite UML tool.

        @param classObj: class information from vppclassdiagram.Class
        @param dictOfInheritance: a dictionary of inheritence info : {'id': vppclassdiagram.Inheritance}
        @return: string with the C++ inheritance for this class, e.g. " : public Parent1, public Parent2"
        for "class Child : public Parent1, public Parent2"
        """
        result = ""
        if isinstance(classObj, Class) and isinstance(dictOfInheritance, dict):
            for id, inheritance in dictOfInheritance.items():
                if isinstance(inheritance, Inheritance):
                    if inheritance.CLASS_TO_ID.find(classObj.ID) > -1:
                        result = result + ' public ' + inheritance.CLASS_FROM.replace(classObj.NAMESPACE + "::","") + ", "
                else:
                    raise RuntimeError("dictOfInheritance contains item not of type 'Inheritance'")
        else:
            raise RuntimeError("classObj not of type 'Class' OR dictOfInheritance not of type 'dict'")

        if len(result) > 0:
            # remove last ','
            result = result.strip().rstrip(',')
            # add " : "
            result = " : " + result
        return result

    # Typically what goes into a H file, when the items can NOT be forward declared in H file.
    def GetNotForwardDeclarableHeaderIncludes(self, classObj, namespace_to_folders = False, filter_out_type_not_in_model = False, include_vector_if_needed = False):
        """
            For Header files. Given a class object, and class diagram, will return a multi-line string containing all header includes
            for types that can not be forward declared (ie. to reduce include dependencies) in a header file.

            If 'namespace_to_folders' is set, then that assumes the code is output into a separate folder per namespace.

            @param classObj: the 'vppclassdiagram.Class' object that we are working out the not-forward-declarable-type header includes.
            @param namespace_to_folders: This option implies that code is generated nested, in folders, named after namespace. This affects how header files are addressed.
            @param filter_out_type_not_in_model:  Will filter out includes of classes that do not exist in the class diagram.
            @return:
        """
        if isinstance(classObj, Class):
            filter = classObj.GetNotForwardDeclarableNonPrimitiveTypesLinkedToThis()
            namespace_to_classes = _getNamespaceToClassesFromFullyQualifiedNames(classObj, filter, True)
            if filter_out_type_not_in_model:
                namespace_to_classes = _filterOutTypesNotInModel(namespace_to_classes, classObj.parent_classDiagram)
            result = _getIncludeStringFromNamespaceToClassMap(namespace_to_classes,namespace_to_folders)

            # Check multiplicity of attributes / operation parameters / associations for 'array'...and include that if it needs.
            if include_vector_if_needed:
                if classObj.DoAttributesAssociationsReturnTypesOrFunctionParametersRequireVector():
                    result = result + '#include <vector>\n'

            return result
        else:
            raise RuntimeError("classObj not of type 'Class'")

    # Typically what goes into a CPP file, when the items can be forward declared in H file.
    def GetForwardDeclarableHeaderIncludes(self, classObj, namespace_to_folders = False, filter_out_type_not_in_model = False):
        """
            For SOURCE files. Given a class object, and class diagram, will return a multi-line string containing all header includes
            for types that can be forward declared (ie. to reduce include dependencies) in a header file.

            If 'namespace_to_folders' is set, then that assumes the code is output into a separate folder per namespace.

            @param classObj: the 'vppclassdiagram.Class' object that we are working out the forward-declarable-type header includes.
            @param namespace_to_folders: This option implies that code is generated nested, in folders, named after namespace. This affects how header files are addressed.
        """
        if isinstance(classObj, Class):
            filter = classObj.GetForwardDeclarableNonPrimitiveTypesLinkedToThis()
            namespace_to_classes = _getNamespaceToClassesFromFullyQualifiedNames(classObj, filter, True)
            if filter_out_type_not_in_model:
                namespace_to_classes = _filterOutTypesNotInModel(namespace_to_classes, classObj.parent_classDiagram)
            result = _getIncludeStringFromNamespaceToClassMap(namespace_to_classes,namespace_to_folders)
            return result
        else:
            raise RuntimeError("classObj not of type 'Class' OR classDiagram not of type 'ClassDiagram'")

    # Typically what goes into a H file, when the items can be forward declared in H file.
    def GetForwardDeclarations(self, classObj):
        """
        For HEADER files. Given a class object, and class diagram, will return a multi-line string containing all forward declares
        for types that can be forward declared (ie. to reduce include dependencies) in a header file.

        @param classObj:
        """
        if isinstance(classObj, Class):
            filter = classObj.GetForwardDeclarableNonPrimitiveTypesLinkedToThis()
            namespace_to_classes = _getNamespaceToClassesFromFullyQualifiedNames(classObj, filter, False)
            result = ""
            for _namespace, _classes in namespace_to_classes.items():
                result = result + _getFormatNestedNamespaceBegin(_namespace) + "\n"
                for _class in _classes:
                    result = result + "    class " + _class + ";\n"
                result = result + _getFormatNestedNamespaceEnd(_namespace) + "\n"

            if result:
                result = "// Begin Forward declarations\n" + result + "// End Forward declarations"
            return result
        else:
            raise RuntimeError("classObj not of type 'Class' OR classDiagram not of type 'ClassDiagram'")

    def GetTypeAndNameFromMultiplicityAndModifier(self, classObj, TYPE, TYPE_MODIFIER, MULTIPLICITY, NAME):
        '''
        https://www.researchgate.net/figure/Mappings-from-C-declarations-to-UML-multiplicity-ranges-depend-on-pointer-reference_tbl1_4207986
        http://www.cs.kent.edu/~jmaletic/papers/JIST07.pdf

        @param TYPE:
        @param TYPE_MODIFIER:
        @param MULTIPLICITY:
        @param NAME:
        @return:
        '''
        if TYPE_MODIFIER.strip() == "[]":
            if not MULTIPLICITY:  # vector
                return ["std::vector<" + TYPE + ">", NAME]
            #else:
            #    TYPE_MODIFIER = ""

        containerType = classObj.GetContainerMultiplicityType(MULTIPLICITY)
        if containerType.find('vector') > -1:
            return ["std::vector<" + TYPE + TYPE_MODIFIER + ">", NAME]
        elif containerType.find('array') > -1:
            return [TYPE + TYPE_MODIFIER, NAME + "[" + containerType.split(":")[-1] + "]"]
        return [TYPE + TYPE_MODIFIER, NAME]

    def GetDefaultFormatFromMultiplicityAndModifier(self, classObj, TYPE_MODIFIER, MULTIPLICITY, DEFAULT):
        if TYPE_MODIFIER.strip() == "[]":
            if not MULTIPLICITY:  # vector
                return " = {" + DEFAULT + "}"

        containerType = classObj.GetContainerMultiplicityType(MULTIPLICITY)
        if containerType.find('vector') > -1:
            return " = {" + DEFAULT + "}"
        elif containerType.find('array') > -1:  # Array -> In C++ it is not possible to pass a complete block of memory by value as a parameter to a function, but we are allowed to pass its address.
            return ""
        return " = " + DEFAULT

    def GetAttributeDeclarationsPerVisibility(self, classObj, visibility="all", attributes=None):
        """
        Get the declaration string for all attributes of the desired visibility
         - public
         - protected
         - private
         - all (public and protected and private)

        @param classObj: vppclassdiagram.Class
        @param visibility: string
        """
        result = ""
        no_attr_passed_in = attributes is None
        if no_attr_passed_in:
            attributes = classObj.ATTRIBUTES

        for attr in attributes:
            if visibility.lower().strip() == attr.VISIBILITY.lower().strip() or visibility.lower().strip() == 'all':
                result = result + self.FormatLongComment(attr.USER_COMMENTS)
                type_name = self.GetTypeAndNameFromMultiplicityAndModifier(classObj, attr.TYPE, attr.TYPE_MODIFIER, attr.MULTIPLICITY, attr.NAME)
                result = result + self.InstantiateType(type_name[0], type_name[1], '', False, attr.IS_STATIC, attr.IS_CONST)
                result = result + ";\n"
        if result and no_attr_passed_in:
            result = "/// @{ " + visibility.capitalize() + " attributes\n" + result + "/// @}"
        return result.rstrip("\n")

    def GetAssociationDeclarationsPerVisibility(self, classObj, visibility="all"):
        """
        Get the declaration string for all associations of the desired visibility
         - public
         - protected
         - private
         - all (public and protected and private)

        @param classObj: vppclassdiagram.Class
        @param visibility: string
        """
        result = self.GetAttributeDeclarationsPerVisibility(classObj, visibility, classObj.GetAssociationsAsListOfAttributesPerVisibility(visibility))
        if result:
            result = "/// @{ " + visibility.capitalize() + " associations\n" + result + "\n/// @}"
        return result.rstrip("\n")

    def GetStaticAttributeDefinitions(self, classObj, attributes=None):
        """
        Get the definition string for all static-attributes regardless of the desired visibility
         - public
         - protected
         - private
         - all (public and protected and private)

        @param classObj: vppclassdiagram.Class
        """
        result = ""
        no_attr_passed_in = attributes is None
        if no_attr_passed_in:
            attributes = classObj.ATTRIBUTES

        for attr in attributes:
            if attr.IS_STATIC:
                type_name = self.GetTypeAndNameFromMultiplicityAndModifier(classObj, attr.TYPE, attr.TYPE_MODIFIER, attr.MULTIPLICITY, attr.NAME)
                result = result + self.InstantiateType(type_name[0], classObj.NAME + "::" + type_name[1], ("" if not attr.INITIAL_VALUE else attr.INITIAL_VALUE), False, False, attr.IS_CONST)
                result = result + ";\n"
        if result and no_attr_passed_in:
            result = "/// @{ Static attribute definitions\n" + result + "/// @}"
        return result.rstrip("\n")

    def GetStaticAssociationDefinitions(self, classObj):
        """
        Get the declaration string for all static-associations regardless of the desired visibility
         - public
         - protected
         - private
         - all (public and protected and private)

        @param classObj: vppclassdiagram.Class
        """
        result = self.GetStaticAttributeDefinitions(classObj, classObj.GetAssociationsAsListOfAttributesPerVisibility("all"))
        if result:
            result = "/// @{ Static association definitions\n" + result + "\n/// @}"
        return result.rstrip("\n")

    def GetOperationPerVisibility(self, classObj, is_impl, visibility="all", REALIZING_CLASS = ""):
        result = ""
        # First see if we realize some base-classes.
        realized_result = ""
        for id, inheritance in classObj.parent_classDiagram.inheritence.items():
            if inheritance.CLASS_TO_ID.find(classObj.ID) > -1:
                if inheritance.IS_REALIZATION or REALIZING_CLASS:  # or REALIZING_CLASS -> if a pure virtual interface inherits another pure virtual interface, and the top child class that inherits them all wants to realize, then all should be realized
                    # get the class we are realizing...
                    realizeObj = classObj.parent_classDiagram.classes[inheritance.CLASS_FROM_ID]
                    if realizeObj.PURE_VIRTUAL_INTERFACE:
                        realized_result = realized_result + self.GetOperationPerVisibility(realizeObj,is_impl,visibility, (classObj.NAME if not REALIZING_CLASS else REALIZING_CLASS)) + "\n"
        if realized_result:
            result = realized_result

        for operation in classObj.OPERATIONS:
            if visibility.lower().strip() == operation.VISIBILITY.lower().strip() or visibility.lower().strip() == 'all':

                is_constructor = operation.NAME.strip() == classObj.NAME.strip()

                result = result + self.FormatLongComment(operation.USER_COMMENTS)
                # Create a list of tuples for the parameters.
                # parameter is a dictionary with the following keys:
                # 'const'
                # 'type'
                # 'name'
                # 'modifier'
                # 'defaultvalue'
                # 'multiplicity'
                params = []
                for param in operation.PARAMETERS:
                    tuples = self.GetTypeAndNameFromMultiplicityAndModifier(classObj, param["type"].strip(), param["modifier"].strip(), param["multiplicity"].strip(), param["name"].strip())
                    tuple0 = param["const"].strip() + " " + tuples[0]
                    tuple0 = tuple0.lstrip()
                    tuple1 = tuples[1]
                    if not is_impl:  # declarations need defaults.
                        tuple1 = tuple1 + ("" if not param["defaultvalue"].strip() else self.GetDefaultFormatFromMultiplicityAndModifier(classObj, param["modifier"].strip(), param["multiplicity"].strip(), param["defaultvalue"]))
                    params.append((tuple0, tuple1))

                # How to handle a return type modifier of [] ??? There is no multiplicity.
                # For now I do it as a vector in this case...
                # Search for <!#!>
                return_type = "" if is_constructor else self.GetTypeAndNameFromMultiplicityAndModifier(classObj, operation.RETURN_TYPE, operation.RETURN_TYPE_MODIFIER, "", "")[0]  # operation.RETURN_TYPE + operation.RETURN_TYPE_MODIFIER

                classname = (classObj.NAME if not REALIZING_CLASS else REALIZING_CLASS)
                result = result + self.DeclareFunction(return_type, classname, operation.NAME, is_impl, params, classObj.PURE_VIRTUAL_INTERFACE or operation.VIRTUAL, operation.IS_STATIC, operation.IS_CONST).lstrip()
                if not is_impl:
                    if classObj.PURE_VIRTUAL_INTERFACE:
                        if not REALIZING_CLASS:
                            result = result + " = 0"
                        else:
                            result = result + " override"
                    result = result + ";\n"
                else:
                    result = result + "\n"
                    result = result + "{\n"
                    result = result + "    /// {{{USER_" + ("CONSTRUCTOR" if is_constructor else CleanName(operation.RETURN_TYPE)) + "_" + classname + "_" + operation.NAME + "_" + str(len(params)) + "_PARAMS}}}\n"
                    result = result + "    /// {{{USER_" + ("CONSTRUCTOR" if is_constructor else CleanName(operation.RETURN_TYPE)) + "_" + classname + "_" + operation.NAME + "_" + str(len(params)) + "_PARAMS}}}\n"
                    result = result + "}\n"

        if result.replace('\n',"").strip():
            if is_impl:
                result = "/// @{ " + visibility.capitalize() + " operation implementations\n" + result + "/// @}" if not REALIZING_CLASS else result
            else:
                result = "/// @{ " + visibility.capitalize() + " operation declarations\n" + result + "/// @}" if not REALIZING_CLASS else "/// @{ " + classObj.NAME + " operation overrides\n" + result + "/// @}"
        return result.rstrip("\n")

    def GetConstructor(self, classObj, is_impl = False):
        # Problem is, that const attributes need to be initialized in a constructor.
        # static const ones are already (and can only be) done where they are declared in CPP file...but class members need to be done in the constructor.
        # This works all that out.
        all = [classObj.ATTRIBUTES, classObj.GetAssociationsAsListOfAttributesPerVisibility("all")]
        params = []
        attr_names = []
        for a in all:
            for attr in a:
                if attr.IS_CONST and not attr.IS_STATIC:
                    tuples = self.GetTypeAndNameFromMultiplicityAndModifier(classObj, attr.TYPE.strip(), attr.TYPE_MODIFIER.strip(), attr.MULTIPLICITY.strip(),  attr.NAME.strip().lower().replace("m_","_"))
                    tuple0 = "const " + tuples[0]
                    tuple0 = tuple0.lstrip()
                    tuple1 = tuples[1]
                    params.append((tuple0,tuple1))
                    attr_names.append(attr.NAME.strip())
        result = ""
        if params:
            result = result + self.DeclareFunction("", classObj.NAME, classObj.NAME, is_impl, params, False, False, False).lstrip() + ("" if is_impl else ";")
            if is_impl:
                result = result + "\n: "
                for i in range(len(params)):
                    result = result + attr_names[i] + "{ " + params[i][1] + " }, "
                result = result.rstrip(", ")
                result = result + "\n"
                result = result + "{\n"
                result = result + "    /// {{{USER_CONSTRUCTOR_" + classObj.NAME + "_" + str(len(params)) + "_PARAMS}}}\n"
                result = result + "    /// {{{USER_CONSTRUCTOR_" + classObj.NAME + "_" + str(len(params)) + "_PARAMS}}}\n"
                result = result + "}\n"
        return result.rstrip("\n")

    def GetAttributeAssociationGetterSetter(self, classObj, is_impl = False):
        result = ""
        # start with attributes, then associations
        all = [classObj.ATTRIBUTES, classObj.GetAssociationsAsListOfAttributesPerVisibility("protected") + classObj.GetAssociationsAsListOfAttributesPerVisibility("private")]
        for a in all:
            for attr in a:
                type_and_name = self.GetTypeAndNameFromMultiplicityAndModifier(classObj, attr.TYPE, attr.TYPE_MODIFIER, attr.MULTIPLICITY, attr.NAME)
                a_type = ("const " if attr.IS_CONST else "") + type_and_name[0] + "&"
                if attr.HAS_GETTER:
                    result = result + self.DeclareFunction(a_type, classObj.NAME, type_and_name[1].replace("m_", "Get"), is_impl) + ("\n" if is_impl else ";\n")
                    if is_impl:
                        result = result + "{\n"
                        result = result + "    return " + type_and_name[1] + ";\n"
                        result = result + "}\n"
                if attr.HAS_SETTER and not attr.IS_CONST:  # Cant set a const attribute!
                    inputvarName = type_and_name[1].replace("m_", "__new")
                    result = result + self.DeclareFunction("void", classObj.NAME, type_and_name[1].replace("m_", "Set"), is_impl, [(a_type, inputvarName)]) + ("\n" if is_impl else ";\n")
                    if is_impl:
                        result = result + "{\n"
                        result = result + "    " + type_and_name[1] + " = " + inputvarName + ";\n"
                        result = result + "}\n"

        if result:
            if is_impl:
                result = "/// @{ Accessor implementations\n" + result + "/// @}"
            else:
                result = "/// @{ Accessor declarations\n" + result + "/// @}"
        return result.rstrip("\n")

    def GetEnumLiterals(self, classObj):
        if not classObj.IS_ENUM:
            raise RuntimeError("Can't get enum literals from a non-Enum")
        result = ""
        cnt = 0
        for literal in classObj.ENUM_LITERALS:
            result = result + literal + " = " + str(cnt) + ",\n"
            cnt = cnt + 1
        result = result.rstrip(",\n")

        return result

    def GetStructMembers(self, classObj):
        if not classObj.IS_STRUCT:
            raise RuntimeError("Can't get struct members from a non-Struct")
        result = ""
        for attr in classObj.ATTRIBUTES:
            result = result + self.FormatLongComment(attr.USER_COMMENTS)
            type_name = self.GetTypeAndNameFromMultiplicityAndModifier(classObj, attr.TYPE, attr.TYPE_MODIFIER, attr.MULTIPLICITY, attr.NAME)
            result = result + self.InstantiateType(type_name[0], type_name[1], '', classObj.IS_STRUCT_PACKED, False, False)
            result = result + ";\n"
        return result

    def GetPacked(self, classObj, is_begin):
        if not classObj.IS_STRUCT:
            raise RuntimeError("Can't get struct members from a non-Struct")
        result = ""
        if classObj.IS_STRUCT_PACKED:
            if is_begin:
                result = result + "#if defined(_MSC_VER)\n"
                result = result + "    #define __attribute__(x)\n"
                result = result + "    #pragma pack(push,1)\n"
                result = result + "#elif defined(__GNUC__)\n"
                result = result + "#elif defined(__clang__)\n"
                result = result + "#endif\n"
            else:
                result = result + "#if defined(_MSC_VER)\n"
                result = result + "    #pragma pack(pop)\n"
                result = result + "#elif defined(__GNUC__) || defined(__clang__)\n"
                result = result + "#endif\n"
        return result

    def GetProjectIncludes(self, setOfProjectDependencies):
        pass # not yet implemented.

def _filterOutTypesNotInModel(namespace_to_classes, classDiagram):
    """
    Will filter out class names in the 'namespace_to_classes' structure that do not exist in the class diagram.

    @param namespace_to_classes: a dictionary of lists - {"XNameSpace1::XNameSpace2":[Class1, Class2, Class3]}
    @param classDiagram: vppclassdiagram.ClassDiagram

    """
    namespace_to_classes_only_in_model = {}
    for ns, classes in namespace_to_classes.items():
        new_classes = []
        for c in classes:
            for id, pc in classDiagram.classes.items():
                if pc.NAME == c:
                    new_classes.append(c)
        if len(new_classes) > 0:
            namespace_to_classes_only_in_model[ns] = new_classes
    return namespace_to_classes_only_in_model


def _getFormatNestedNamespaceBegin(fullyQualifiedNamespace):
    """
    Returns the C++ format string for a nested namespace declaration on one line...
    @param fullyQualifiedNamespace: a string i.e. "XNamespace1::XNamespace2::CClass"
    @return: string i.e. "namespace XFullStack { namespace XProtocol {"
    """
    # Nested NAMESPACE of the classobj is i.e. 'XFullStack::XProtocol'
    nestedNS = fullyQualifiedNamespace.split("::")
    result = ""
    for n in nestedNS:
         result = result + " namespace " + n + " { "
    result = result.lstrip(" ")
    return result


def _getFormatNestedNamespaceEnd(fullyQualifiedNamespace):
    """
    Returns the C++ format string for a nested namespace end-declaration on one line...
    @param fullyQualifiedNamespace: a string i.e. "XNamespace1::XNamespace2::CClass"
    @return: string i.e. for "namespace XFullStack { namespace XProtocol {" will return "} }"
    """
    # Nested NAMESPACE of the classobj is i.e. 'XFullStack::XProtocol'
    nestedNS = fullyQualifiedNamespace.split("::")
    result = ""
    for n in nestedNS:
        result = result + " } "
    result = result.lstrip()
    return result


def _getIncludeStringFromNamespaceToClassMap(namespace_to_class, namespace_to_folders):
    result = ""
    for path, classnames in namespace_to_class.items():
        for classname in classnames:
            if not namespace_to_folders:  # clear the path
                path = ""
            result = result + '#include "' + path + classname + ".h" + '"\n'
    return result


def _getNamespaceToClassesFromFullyQualifiedNames(classObj, setOfClasses, is_file_include):
    """
    In the ClassDiagramModel, all Types are full qualified, ie 'XFullStack::XProtocol::CProtocolStack'
    It is thus safe to assume that should one split on ::, if the result is one, there is no namespace.

    @param setOfClasses: Set of all classes extracted for classObj
    @param classObj:
    @return: dictionary {'NS1::NS2':'CClass'}
    """
    namespace_to_class = OrderedDict()
    # Get a dictionary of namespace names, with all classes within.
    for f in setOfClasses:
        if is_file_include:
            # If a class is in the same namespace...then it is in the same folder...so clean this first.
            f = f.replace(classObj.NAMESPACE + "::", "")
        full = f.split("::")
        ns = f.replace(full[-1], "")
        if is_file_include:  # when using this for include files, switch :: with /
            ns = ns.replace("::", "/")
        else:  # when using this for forward declarations, the extra '::' creates an empty namespace.
            ns = ns.rstrip("::")
        if ns not in namespace_to_class:
            namespace_to_class[ns] = []
        namespace_to_class[ns].append(full[-1])
    return namespace_to_class

